from unittest import IsolatedAsyncioTestCase

from aiohttp.web import HTTPForbidden

from ...admin.request_context import AdminRequestContext
from ...ledger.base import BaseLedger
from ...protocols.coordinate_mediation.v1_0.route_manager import RouteManager
from ...tests import mock
from ...utils.testing import create_test_profile
from ...wallet.did_method import SOV, DIDMethod, DIDMethods, HolderDefinedDid
from ...wallet.key_type import ED25519, KeyTypes
from .. import routes as test_module
from ..base import BaseWallet
from ..did_info import DIDInfo
from ..did_posture import DIDPosture

WEB = DIDMethod(
    name="web",
    key_types=[ED25519],
    rotation=True,
    holder_defined_did=HolderDefinedDid.REQUIRED,
)


class TestWalletRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.wallet = mock.create_autospec(BaseWallet)
        self.session_inject = {BaseWallet: self.wallet}
        self.profile = await create_test_profile(
            settings={"admin.admin_api_key": "secret-key"},
        )

        self.route_mgr = mock.MagicMock(RouteManager, autospec=True)
        self.route_mgr.mediation_record_if_id = mock.CoroutineMock(return_value=None)
        self.route_mgr.routing_info = mock.CoroutineMock(return_value=(None, None))
        self.profile.context.injector.bind_instance(RouteManager, self.route_mgr)
        self.profile.context.injector.bind_instance(KeyTypes, KeyTypes())
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

        self.test_did = "did"
        self.test_did_sov = "WgWxqztrNooG92RXvxSTWv"
        self.test_did_web = "did:web:doma.in"
        self.test_verkey = "verkey"
        self.test_posted_did = "posted-did"
        self.test_posted_verkey = "posted-verkey"
        self.did_methods = DIDMethods()
        self.did_methods.register(WEB)
        self.context.injector.bind_instance(DIDMethods, self.did_methods)

        self.test_mediator_routing_keys = ["3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRR"]
        self.test_mediator_endpoint = "http://mediator.example.com"

    async def test_missing_wallet(self):
        self.session_inject[BaseWallet] = None

        with self.assertRaises(HTTPForbidden):
            await test_module.wallet_create_did(self.request)

        with self.assertRaises(HTTPForbidden):
            await test_module.wallet_did_list(self.request)

        with self.assertRaises(HTTPForbidden):
            await test_module.wallet_get_public_did(self.request)

        with self.assertRaises(HTTPForbidden):
            await test_module.wallet_set_public_did(self.request)

        with self.assertRaises(HTTPForbidden):
            self.request.json = mock.CoroutineMock(
                return_value={
                    "did": self.test_did,
                    "endpoint": "https://my-endpoint.ca:8020",
                }
            )
            await test_module.wallet_set_did_endpoint(self.request)

        with self.assertRaises(HTTPForbidden):
            await test_module.wallet_get_did_endpoint(self.request)

    def test_format_did_info(self):
        did_info = DIDInfo(
            self.test_did,
            self.test_verkey,
            DIDPosture.WALLET_ONLY.metadata,
            SOV,
            ED25519,
        )
        result = test_module.format_did_info(did_info)
        assert (
            result["did"] == self.test_did
            and result["verkey"] == self.test_verkey
            and result["posture"] == DIDPosture.WALLET_ONLY.moniker
        )

        did_info = DIDInfo(
            self.test_did,
            self.test_verkey,
            {"posted": True, "public": True},
            SOV,
            ED25519,
        )
        result = test_module.format_did_info(did_info)
        assert result["posture"] == DIDPosture.PUBLIC.moniker

        did_info = DIDInfo(
            self.test_did,
            self.test_verkey,
            {"posted": True, "public": False},
            SOV,
            ED25519,
        )
        result = test_module.format_did_info(did_info)
        assert result["posture"] == DIDPosture.POSTED.moniker

    async def test_create_did(self):
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.wallet.create_local_did.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                DIDPosture.WALLET_ONLY.metadata,
                SOV,
                ED25519,
            )
            result = await test_module.wallet_create_did(self.request)
            json_response.assert_called_once_with(
                {
                    "result": {
                        "did": self.test_did,
                        "verkey": self.test_verkey,
                        "posture": DIDPosture.WALLET_ONLY.moniker,
                        "key_type": ED25519.key_type,
                        "method": SOV.method_name,
                        "metadata": {"posted": False, "public": False},
                    }
                }
            )
            assert result is json_response.return_value

    async def test_create_did_unsupported_method(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "method": "madeupmethod",
                "options": {"key_type": "bls12381g2"},
            }
        )

        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.wallet_create_did(self.request)

    async def test_create_did_unsupported_key_type(self):
        self.request.json = mock.CoroutineMock(
            return_value={"method": "sov", "options": {"key_type": "bls12381g2"}}
        )
        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.wallet_create_did(self.request)

    async def test_create_did_indy(self):
        self.request.json = mock.CoroutineMock(
            return_value={"method": "indy", "options": {"key_type": ED25519.key_type}}
        )
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.wallet_create_did(self.request)

    async def test_create_did_method_requires_user_defined_did(self):
        # given
        did_custom = DIDMethod(
            name="custom",
            key_types=[ED25519],
            rotation=True,
            holder_defined_did=HolderDefinedDid.REQUIRED,
        )
        self.did_methods.register(did_custom)

        self.request.json = mock.CoroutineMock(
            return_value={"method": "custom", "options": {"key_type": "ed25519"}}
        )

        # when - then
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.wallet_create_did(self.request)

    async def test_create_did_method_doesnt_support_user_defined_did(self):
        did_custom = DIDMethod(
            name="custom",
            key_types=[ED25519],
            rotation=True,
            holder_defined_did=HolderDefinedDid.NO,
        )
        self.did_methods.register(did_custom)

        # when
        self.request.json = mock.CoroutineMock(
            return_value={
                "method": "custom",
                "options": {
                    "key_type": ED25519.key_type,
                    "did": "did:custom:aCustomUserDefinedDID",
                },
            }
        )

        # then
        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.wallet_create_did(self.request)

    async def test_create_did_x(self):
        self.wallet.create_local_did.side_effect = test_module.WalletError()
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.wallet_create_did(self.request)

    async def test_did_list(self):
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:  # , mock.patch.object(
            self.wallet.get_local_dids.return_value = [
                DIDInfo(
                    self.test_did,
                    self.test_verkey,
                    DIDPosture.WALLET_ONLY.metadata,
                    SOV,
                    ED25519,
                ),
                DIDInfo(
                    self.test_posted_did,
                    self.test_posted_verkey,
                    DIDPosture.POSTED.metadata,
                    SOV,
                    ED25519,
                ),
            ]
            result = await test_module.wallet_did_list(self.request)
            json_response.assert_called_once_with(
                {
                    "results": [
                        {
                            "did": self.test_posted_did,
                            "verkey": self.test_posted_verkey,
                            "posture": DIDPosture.POSTED.moniker,
                            "key_type": ED25519.key_type,
                            "method": SOV.method_name,
                            "metadata": {"posted": True, "public": False},
                        },
                        {
                            "did": self.test_did,
                            "verkey": self.test_verkey,
                            "posture": DIDPosture.WALLET_ONLY.moniker,
                            "key_type": ED25519.key_type,
                            "method": SOV.method_name,
                            "metadata": {"posted": False, "public": False},
                        },
                    ]
                }
            )
            assert json_response.return_value is json_response()
            assert result is json_response.return_value

    async def test_did_list_filter_public(self):
        self.request.query = {"posture": DIDPosture.PUBLIC.moniker}
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.wallet.get_public_did.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                DIDPosture.PUBLIC.metadata,
                SOV,
                ED25519,
            )
            self.wallet.get_posted_dids.return_value = [
                DIDInfo(
                    self.test_posted_did,
                    self.test_posted_verkey,
                    DIDPosture.POSTED.metadata,
                    SOV,
                    ED25519,
                )
            ]
            result = await test_module.wallet_did_list(self.request)
            json_response.assert_called_once_with(
                {
                    "results": [
                        {
                            "did": self.test_did,
                            "verkey": self.test_verkey,
                            "posture": DIDPosture.PUBLIC.moniker,
                            "key_type": ED25519.key_type,
                            "method": SOV.method_name,
                            "metadata": {"posted": True, "public": True},
                        }
                    ]
                }
            )
            assert json_response.return_value is json_response()
            assert result is json_response.return_value

    async def test_did_list_filter_posted(self):
        self.request.query = {"posture": DIDPosture.POSTED.moniker}
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.wallet.get_public_did.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                DIDPosture.PUBLIC.metadata,
                SOV,
                ED25519,
            )
            self.wallet.get_posted_dids.return_value = [
                DIDInfo(
                    self.test_posted_did,
                    self.test_posted_verkey,
                    {
                        "posted": True,
                        "public": False,
                    },
                    SOV,
                    ED25519,
                )
            ]
            result = await test_module.wallet_did_list(self.request)
            json_response.assert_called_once_with(
                {
                    "results": [
                        {
                            "did": self.test_posted_did,
                            "verkey": self.test_posted_verkey,
                            "posture": DIDPosture.POSTED.moniker,
                            "key_type": ED25519.key_type,
                            "method": SOV.method_name,
                            "metadata": {"posted": True, "public": False},
                        }
                    ]
                }
            )
            assert json_response.return_value is json_response()
            assert result is json_response.return_value

    async def test_did_list_filter_did(self):
        self.request.query = {"did": self.test_did}
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.wallet.get_local_did.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                DIDPosture.WALLET_ONLY.metadata,
                SOV,
                ED25519,
            )
            result = await test_module.wallet_did_list(self.request)
            json_response.assert_called_once_with(
                {
                    "results": [
                        {
                            "did": self.test_did,
                            "verkey": self.test_verkey,
                            "posture": DIDPosture.WALLET_ONLY.moniker,
                            "key_type": ED25519.key_type,
                            "method": SOV.method_name,
                            "metadata": {"posted": False, "public": False},
                        }
                    ]
                }
            )
            assert json_response.return_value is json_response()
            assert result is json_response.return_value

    async def test_did_list_filter_did_x(self):
        self.request.query = {"did": self.test_did}
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.wallet.get_local_did.side_effect = test_module.WalletError()
            result = await test_module.wallet_did_list(self.request)
            json_response.assert_called_once_with({"results": []})
            assert json_response.return_value is json_response()
            assert result is json_response.return_value

    async def test_did_list_filter_verkey(self):
        self.request.query = {"verkey": self.test_verkey}
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.wallet.get_local_did_for_verkey.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                DIDPosture.WALLET_ONLY.metadata,
                SOV,
                ED25519,
            )
            result = await test_module.wallet_did_list(self.request)
            json_response.assert_called_once_with(
                {
                    "results": [
                        {
                            "did": self.test_did,
                            "verkey": self.test_verkey,
                            "posture": DIDPosture.WALLET_ONLY.moniker,
                            "key_type": ED25519.key_type,
                            "method": SOV.method_name,
                            "metadata": {"posted": False, "public": False},
                        }
                    ]
                }
            )
            assert json_response.return_value is json_response()
            assert result is json_response.return_value

    async def test_did_list_filter_verkey_x(self):
        self.request.query = {"verkey": self.test_verkey}
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.wallet.get_local_did_for_verkey.side_effect = test_module.WalletError()
            result = await test_module.wallet_did_list(self.request)
            json_response.assert_called_once_with({"results": []})
            assert json_response.return_value is json_response()
            assert result is json_response.return_value

    async def test_get_public_did(self):
        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.wallet.get_public_did.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                DIDPosture.PUBLIC.metadata,
                SOV,
                ED25519,
            )
            result = await test_module.wallet_get_public_did(self.request)
            json_response.assert_called_once_with(
                {
                    "result": {
                        "did": self.test_did,
                        "verkey": self.test_verkey,
                        "posture": DIDPosture.PUBLIC.moniker,
                        "key_type": ED25519.key_type,
                        "method": SOV.method_name,
                        "metadata": {"posted": True, "public": True},
                    }
                }
            )
            assert result is json_response.return_value

    async def test_get_public_did_x(self):
        self.wallet.get_public_did.side_effect = test_module.WalletError()
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.wallet_get_public_did(self.request)

    async def test_set_public_did(self):
        self.request.query = {"did": self.test_did}

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.get_key_for_did = mock.CoroutineMock()
        ledger.update_endpoint_for_did = mock.CoroutineMock()
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        mock_route_manager = mock.MagicMock(RouteManager, autospec=True)
        mock_route_manager.route_verkey = mock.CoroutineMock()
        mock_route_manager.mediation_record_if_id = mock.CoroutineMock()
        mock_route_manager.routing_info = mock.CoroutineMock(
            return_value=(self.test_mediator_routing_keys, self.test_mediator_endpoint)
        )
        self.profile.context.injector.bind_instance(RouteManager, mock_route_manager)

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.wallet.set_public_did.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                DIDPosture.PUBLIC.metadata,
                SOV,
                ED25519,
            )
            self.wallet.get_local_did = mock.CoroutineMock()
            self.wallet.set_did_endpoint = mock.CoroutineMock()
            result = await test_module.wallet_set_public_did(self.request)
            self.wallet.set_public_did.assert_awaited_once()
            json_response.assert_called_once_with(
                {
                    "result": {
                        "did": self.test_did,
                        "verkey": self.test_verkey,
                        "posture": DIDPosture.PUBLIC.moniker,
                        "key_type": ED25519.key_type,
                        "method": SOV.method_name,
                        "metadata": {"posted": True, "public": True},
                    }
                }
            )
            assert result is json_response.return_value

    async def test_set_public_did_no_query_did(self):
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.wallet_set_public_did(self.request)

    async def test_set_public_did_no_ledger(self):
        mock_route_manager = mock.MagicMock(RouteManager, autospec=True)
        mock_route_manager.mediation_record_if_id = mock.CoroutineMock()
        mock_route_manager.routing_info = mock.CoroutineMock(
            return_value=(self.test_mediator_routing_keys, self.test_mediator_endpoint)
        )
        self.profile.context.injector.bind_instance(RouteManager, mock_route_manager)
        self.request.query = {"did": self.test_did_sov}

        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.wallet_set_public_did(self.request)

    async def test_set_public_did_not_public(self):
        self.request.query = {"did": self.test_did_sov}

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.get_key_for_did = mock.CoroutineMock(return_value=None)
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        mock_route_manager = mock.MagicMock(RouteManager, autospec=True)
        mock_route_manager.mediation_record_if_id = mock.CoroutineMock()
        mock_route_manager.routing_info = mock.CoroutineMock(
            return_value=(self.test_mediator_routing_keys, self.test_mediator_endpoint)
        )
        self.profile.context.injector.bind_instance(RouteManager, mock_route_manager)

        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.wallet_set_public_did(self.request)

    async def test_set_public_did_not_found(self):
        self.request.query = {"did": self.test_did_sov}

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.get_key_for_did = mock.CoroutineMock(return_value=None)
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        mock_route_manager = mock.MagicMock(RouteManager, autospec=True)
        mock_route_manager.mediation_record_if_id = mock.CoroutineMock()
        mock_route_manager.routing_info = mock.CoroutineMock(
            return_value=(self.test_mediator_routing_keys, self.test_mediator_endpoint)
        )
        self.profile.context.injector.bind_instance(RouteManager, mock_route_manager)

        self.wallet.get_local_did.side_effect = test_module.WalletNotFoundError()
        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.wallet_set_public_did(self.request)

    async def test_set_public_did_x(self):
        self.request.query = {"did": self.test_did_sov}

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.update_endpoint_for_did = mock.CoroutineMock()
        ledger.get_key_for_did = mock.CoroutineMock()
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        mock_route_manager = mock.MagicMock(RouteManager, autospec=True)
        mock_route_manager.mediation_record_if_id = mock.CoroutineMock()
        mock_route_manager.routing_info = mock.CoroutineMock(
            return_value=(self.test_mediator_routing_keys, self.test_mediator_endpoint)
        )
        self.profile.context.injector.bind_instance(RouteManager, mock_route_manager)

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.wallet.get_public_did.return_value = DIDInfo(
                self.test_did_sov,
                self.test_verkey,
                DIDPosture.PUBLIC.metadata,
                SOV,
                ED25519,
            )
            self.wallet.set_public_did.side_effect = test_module.WalletError()
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.wallet_set_public_did(self.request)

    async def test_set_public_did_no_wallet_did(self):
        self.request.query = {"did": self.test_did_sov}

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.update_endpoint_for_did = mock.CoroutineMock()
        ledger.get_key_for_did = mock.CoroutineMock()
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        mock_route_manager = mock.MagicMock(RouteManager, autospec=True)
        mock_route_manager.mediation_record_if_id = mock.CoroutineMock()
        mock_route_manager.routing_info = mock.CoroutineMock(
            return_value=(self.test_mediator_routing_keys, self.test_mediator_endpoint)
        )
        self.profile.context.injector.bind_instance(RouteManager, mock_route_manager)

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.wallet.get_public_did.return_value = DIDInfo(
                self.test_did_sov,
                self.test_verkey,
                DIDPosture.PUBLIC.metadata,
                SOV,
                ED25519,
            )
            self.wallet.set_public_did.side_effect = test_module.WalletNotFoundError()
            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.wallet_set_public_did(self.request)

    async def test_set_public_did_update_endpoint(self):
        self.request.query = {"did": self.test_did_sov}

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.update_endpoint_for_did = mock.CoroutineMock()
        ledger.get_key_for_did = mock.CoroutineMock()
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        mock_route_manager = mock.MagicMock(RouteManager, autospec=True)
        mock_route_manager.route_verkey = mock.CoroutineMock()
        mock_route_manager.mediation_record_if_id = mock.CoroutineMock()
        mock_route_manager.routing_info = mock.CoroutineMock(
            return_value=(self.test_mediator_routing_keys, self.test_mediator_endpoint)
        )
        self.profile.context.injector.bind_instance(RouteManager, mock_route_manager)

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.wallet.set_public_did.return_value = DIDInfo(
                self.test_did_sov,
                self.test_verkey,
                {**DIDPosture.PUBLIC.metadata, "endpoint": self.test_mediator_endpoint},
                SOV,
                ED25519,
            )
            self.wallet.get_local_did = mock.CoroutineMock()
            self.wallet.set_did_endpoint = mock.CoroutineMock()
            result = await test_module.wallet_set_public_did(self.request)
            self.wallet.set_public_did.assert_awaited_once()
            json_response.assert_called_once_with(
                {
                    "result": {
                        "did": self.test_did_sov,
                        "verkey": self.test_verkey,
                        "posture": DIDPosture.PUBLIC.moniker,
                        "key_type": ED25519.key_type,
                        "method": SOV.method_name,
                        "metadata": {
                            "posted": True,
                            "public": True,
                            "endpoint": self.test_mediator_endpoint,
                        },
                    }
                }
            )
            assert result is json_response.return_value

    async def test_set_public_did_update_endpoint_use_default_update_in_wallet(self):
        self.request.query = {"did": self.test_did_sov}
        default_endpoint = "https://default_endpoint.com"
        self.context.update_settings({"default_endpoint": default_endpoint})

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.update_endpoint_for_did = mock.CoroutineMock()
        ledger.get_key_for_did = mock.CoroutineMock()
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        mock_route_manager = mock.MagicMock(RouteManager, autospec=True)
        mock_route_manager.route_verkey = mock.CoroutineMock()
        mock_route_manager.mediation_record_if_id = mock.CoroutineMock(return_value=None)
        mock_route_manager.routing_info = mock.CoroutineMock(return_value=(None, None))
        self.profile.context.injector.bind_instance(RouteManager, mock_route_manager)

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            did_info = DIDInfo(
                self.test_did_sov,
                self.test_verkey,
                DIDPosture.PUBLIC.metadata,
                SOV,
                ED25519,
            )
            self.wallet.get_local_did.return_value = did_info
            self.wallet.set_public_did.return_value = did_info
            result = await test_module.wallet_set_public_did(self.request)
            self.wallet.set_public_did.assert_awaited_once()
            self.wallet.set_did_endpoint.assert_awaited_once_with(
                did_info.did,
                "https://default_endpoint.com",
                ledger,
                write_ledger=True,
                endorser_did=None,
                routing_keys=None,
            )
            json_response.assert_called_once_with(
                {
                    "result": {
                        "did": self.test_did_sov,
                        "verkey": self.test_verkey,
                        "posture": DIDPosture.PUBLIC.moniker,
                        "key_type": ED25519.key_type,
                        "method": SOV.method_name,
                        "metadata": {
                            "posted": True,
                            "public": True,
                        },
                    }
                }
            )
            assert result is json_response.return_value

    async def test_set_public_did_with_non_sov_did(self):
        self.request.query = {"did": self.test_did_web}

        mock_route_manager = mock.MagicMock(RouteManager, autospec=True)
        mock_route_manager.route_verkey = mock.CoroutineMock()
        mock_route_manager.mediation_record_if_id = mock.CoroutineMock()
        mock_route_manager.routing_info = mock.CoroutineMock(
            return_value=(self.test_mediator_routing_keys, self.test_mediator_endpoint)
        )
        self.profile.context.injector.bind_instance(RouteManager, mock_route_manager)

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.wallet.set_public_did.return_value = DIDInfo(
                self.test_did_web,
                self.test_verkey,
                DIDPosture.PUBLIC.metadata,
                WEB,
                ED25519,
            )
            self.wallet.get_local_did = mock.CoroutineMock()
            self.wallet.set_did_endpoint = mock.CoroutineMock()
            result = await test_module.wallet_set_public_did(self.request)
            self.wallet.set_public_did.assert_awaited_once()
            self.wallet.set_did_endpoint.assert_not_called()

            json_response.assert_called_once_with(
                {
                    "result": {
                        "did": self.test_did_web,
                        "verkey": self.test_verkey,
                        "posture": DIDPosture.PUBLIC.moniker,
                        "key_type": ED25519.key_type,
                        "method": WEB.method_name,
                        "metadata": {"posted": True, "public": True},
                    }
                }
            )
            assert result is json_response.return_value

    async def test_promote_wallet_public_did(self):
        # Test successful promotion of Indy DID
        did_info = DIDInfo(
            self.test_did_sov,
            self.test_verkey,
            DIDPosture.WALLET_ONLY.metadata,
            SOV,
            ED25519,
        )

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.get_key_for_did = mock.CoroutineMock(return_value=self.test_verkey)
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        mock_route_manager = mock.MagicMock(RouteManager, autospec=True)
        mock_route_manager.route_verkey = mock.CoroutineMock()
        mock_route_manager.mediation_record_if_id = mock.CoroutineMock()
        mock_route_manager.routing_info = mock.CoroutineMock(
            return_value=(self.test_mediator_routing_keys, self.test_mediator_endpoint)
        )
        self.profile.context.injector.bind_instance(RouteManager, mock_route_manager)

        self.wallet.get_local_did.return_value = did_info
        self.wallet.set_public_did.return_value = DIDInfo(
            self.test_did_sov,
            self.test_verkey,
            DIDPosture.PUBLIC.metadata,
            SOV,
            ED25519,
        )

        result, attrib_def = await test_module.promote_wallet_public_did(
            self.context, self.test_did_sov, write_ledger=True
        )

        assert result.did == self.test_did_sov
        assert result.verkey == self.test_verkey
        assert result.metadata == DIDPosture.PUBLIC.metadata
        self.wallet.set_public_did.assert_called_once()
        mock_route_manager.route_verkey.assert_called_once()

    async def test_promote_wallet_public_did_no_ledger(self):
        # Test promotion attempt without ledger
        with self.assertRaises(PermissionError):
            await test_module.promote_wallet_public_did(
                self.context, self.test_did_sov, write_ledger=True
            )

    async def test_promote_wallet_public_did_not_on_ledger(self):
        # Test promotion of DID not on ledger
        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.get_key_for_did = mock.CoroutineMock(return_value=None)
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        with self.assertRaises(LookupError):
            await test_module.promote_wallet_public_did(
                self.context, self.test_did_sov, write_ledger=True
            )

    async def test_promote_wallet_public_did_with_endorser(self):
        # Test promotion with endorser
        did_info = DIDInfo(
            self.test_did_sov,
            self.test_verkey,
            DIDPosture.WALLET_ONLY.metadata,
            SOV,
            ED25519,
        )

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.get_key_for_did = mock.CoroutineMock(return_value=self.test_verkey)
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        # Mock connection record with endorser info
        connection_record = mock.MagicMock()
        connection_record.metadata_get = mock.CoroutineMock(
            return_value={"endorser_did": "endorser-did"}
        )

        with mock.patch.object(
            test_module.ConnRecord,
            "retrieve_by_id",
            mock.CoroutineMock(return_value=connection_record),
        ):
            self.wallet.get_local_did.return_value = did_info
            self.wallet.set_public_did.return_value = DIDInfo(
                self.test_did_sov,
                self.test_verkey,
                DIDPosture.PUBLIC.metadata,
                SOV,
                ED25519,
            )

            result, attrib_def = await test_module.promote_wallet_public_did(
                self.context,
                self.test_did_sov,
                write_ledger=False,
                connection_id="test-connection-id",
            )

            assert result.did == self.test_did_sov
            assert result.verkey == self.test_verkey
            assert result.metadata == DIDPosture.PUBLIC.metadata

    async def test_promote_wallet_public_did_with_endpoint(self):
        # Test promotion with endpoint update
        did_info = DIDInfo(
            self.test_did_sov,
            self.test_verkey,
            DIDPosture.WALLET_ONLY.metadata,
            SOV,
            ED25519,
        )

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.get_key_for_did = mock.CoroutineMock(return_value=self.test_verkey)
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        mock_route_manager = mock.MagicMock(RouteManager, autospec=True)
        mock_route_manager.route_verkey = mock.CoroutineMock()
        mock_route_manager.mediation_record_if_id = mock.CoroutineMock()
        mock_route_manager.routing_info = mock.CoroutineMock(
            return_value=(self.test_mediator_routing_keys, self.test_mediator_endpoint)
        )
        self.profile.context.injector.bind_instance(RouteManager, mock_route_manager)

        self.wallet.get_local_did.return_value = did_info
        self.wallet.set_public_did.return_value = DIDInfo(
            self.test_did_sov,
            self.test_verkey,
            DIDPosture.PUBLIC.metadata,
            SOV,
            ED25519,
        )
        self.wallet.set_did_endpoint = mock.CoroutineMock()

        result, attrib_def = await test_module.promote_wallet_public_did(
            self.context,
            self.test_did_sov,
            write_ledger=True,
            mediator_endpoint="https://custom-endpoint.com",
        )

        assert result.did == self.test_did_sov
        self.wallet.set_did_endpoint.assert_called_once()

    async def test_promote_wallet_public_did_non_indy(self):
        # Test promotion of non-Indy DID
        did_info = DIDInfo(
            self.test_did_web,
            self.test_verkey,
            DIDPosture.WALLET_ONLY.metadata,
            WEB,
            ED25519,
        )

        mock_route_manager = mock.MagicMock(RouteManager, autospec=True)
        mock_route_manager.route_verkey = mock.CoroutineMock()
        self.profile.context.injector.bind_instance(RouteManager, mock_route_manager)

        self.wallet.get_local_did.return_value = did_info
        self.wallet.set_public_did.return_value = DIDInfo(
            self.test_did_web,
            self.test_verkey,
            DIDPosture.PUBLIC.metadata,
            WEB,
            ED25519,
        )

        result, attrib_def = await test_module.promote_wallet_public_did(
            self.context, self.test_did_web
        )

        assert result.did == self.test_did_web
        assert result.verkey == self.test_verkey
        assert result.metadata == DIDPosture.PUBLIC.metadata
        self.wallet.set_public_did.assert_called_once()
        mock_route_manager.route_verkey.assert_called_once()

    async def test_promote_wallet_public_did_missing_connection(self):
        # Test promotion with missing connection
        did_info = DIDInfo(
            self.test_did_sov,
            self.test_verkey,
            DIDPosture.WALLET_ONLY.metadata,
            SOV,
            ED25519,
        )

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.get_key_for_did = mock.CoroutineMock(return_value=self.test_verkey)
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        with mock.patch.object(
            test_module.ConnRecord,
            "retrieve_by_id",
            mock.CoroutineMock(side_effect=test_module.StorageNotFoundError()),
        ):
            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.promote_wallet_public_did(
                    self.context,
                    self.test_did_sov,
                    write_ledger=False,
                    connection_id="test-connection-id",
                )

    async def test_promote_wallet_public_did_missing_endorser_info(self):
        # Test promotion with missing endorser info
        did_info = DIDInfo(
            self.test_did_sov,
            self.test_verkey,
            DIDPosture.WALLET_ONLY.metadata,
            SOV,
            ED25519,
        )

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.get_key_for_did = mock.CoroutineMock(return_value=self.test_verkey)
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        connection_record = mock.MagicMock()
        connection_record.metadata_get = mock.CoroutineMock(return_value={})

        with mock.patch.object(
            test_module.ConnRecord,
            "retrieve_by_id",
            mock.CoroutineMock(return_value=connection_record),
        ):
            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.promote_wallet_public_did(
                    self.context,
                    self.test_did_sov,
                    write_ledger=False,
                    connection_id="test-connection-id",
                )

    async def test_set_did_endpoint(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "did": self.test_did,
                "endpoint": "https://my-endpoint.ca:8020",
            }
        )

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.update_endpoint_for_did = mock.CoroutineMock()
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        self.wallet.get_local_did.return_value = DIDInfo(
            self.test_did,
            self.test_verkey,
            {"public": False, "endpoint": "http://old-endpoint.ca"},
            SOV,
            ED25519,
        )
        self.wallet.get_public_did.return_value = DIDInfo(
            self.test_did,
            self.test_verkey,
            DIDPosture.PUBLIC.metadata,
            SOV,
            ED25519,
        )

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            await test_module.wallet_set_did_endpoint(self.request)
            json_response.assert_called_once_with({})

    async def test_set_did_endpoint_public_did_no_ledger(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "did": self.test_did,
                "endpoint": "https://my-endpoint.ca:8020",
            }
        )

        self.wallet.get_local_did.return_value = DIDInfo(
            self.test_did,
            self.test_verkey,
            {"public": False, "endpoint": "http://old-endpoint.ca"},
            SOV,
            ED25519,
        )
        self.wallet.get_public_did.return_value = DIDInfo(
            self.test_did,
            self.test_verkey,
            DIDPosture.PUBLIC.metadata,
            SOV,
            ED25519,
        )
        self.wallet.set_did_endpoint.side_effect = test_module.LedgerConfigError()

        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.wallet_set_did_endpoint(self.request)

    async def test_set_did_endpoint_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "did": self.test_did,
                "endpoint": "https://my-endpoint.ca:8020",
            }
        )

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.update_endpoint_for_did = mock.CoroutineMock()
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        self.wallet.set_did_endpoint.side_effect = test_module.WalletError()

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.wallet_set_did_endpoint(self.request)

    async def test_set_did_endpoint_no_wallet_did(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "did": self.test_did,
                "endpoint": "https://my-endpoint.ca:8020",
            }
        )

        ledger = mock.MagicMock(BaseLedger, autospec=True)
        ledger.update_endpoint_for_did = mock.CoroutineMock()
        self.profile.context.injector.bind_instance(BaseLedger, ledger)

        self.wallet.set_did_endpoint.side_effect = test_module.WalletNotFoundError()

        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.wallet_set_did_endpoint(self.request)

    async def test_get_did_endpoint(self):
        self.request.query = {"did": self.test_did}

        self.wallet.get_local_did.return_value = DIDInfo(
            self.test_did,
            self.test_verkey,
            {"public": False, "endpoint": "http://old-endpoint.ca"},
            SOV,
            ED25519,
        )

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            await test_module.wallet_get_did_endpoint(self.request)
            json_response.assert_called_once_with(
                {
                    "did": self.test_did,
                    "endpoint": self.wallet.get_local_did.return_value.metadata[
                        "endpoint"
                    ],
                }
            )

    async def test_get_did_endpoint_no_did(self):
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.wallet_get_did_endpoint(self.request)

    async def test_get_did_endpoint_no_wallet_did(self):
        self.request.query = {"did": self.test_did}

        self.wallet.get_local_did.side_effect = test_module.WalletNotFoundError()

        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.wallet_get_did_endpoint(self.request)

    async def test_get_did_endpoint_wallet_x(self):
        self.request.query = {"did": self.test_did}

        self.wallet.get_local_did.side_effect = test_module.WalletError()

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.wallet_get_did_endpoint(self.request)

    async def test_rotate_did_keypair(self):
        self.request.query = {"did": "did"}

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            self.wallet.get_local_did = mock.CoroutineMock(
                return_value=DIDInfo(
                    "did",
                    "verkey",
                    {"public": False},
                    SOV,
                    ED25519,
                )
            )
            self.wallet.rotate_did_keypair_start = mock.CoroutineMock()
            self.wallet.rotate_did_keypair_apply = mock.CoroutineMock()

            await test_module.wallet_rotate_did_keypair(self.request)
            json_response.assert_called_once_with({})

    async def test_rotate_did_keypair_missing_wallet(self):
        self.request.query = {"did": "did"}
        self.session_inject[BaseWallet] = None

        with self.assertRaises(HTTPForbidden):
            await test_module.wallet_rotate_did_keypair(self.request)

    async def test_rotate_did_keypair_no_query_did(self):
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.wallet_rotate_did_keypair(self.request)

    async def test_rotate_did_keypair_did_not_local(self):
        self.request.query = {"did": "did"}

        self.wallet.get_local_did = mock.CoroutineMock(
            side_effect=test_module.WalletNotFoundError("Unknown DID")
        )
        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.wallet_rotate_did_keypair(self.request)

        self.wallet.get_local_did = mock.CoroutineMock(
            return_value=DIDInfo(
                "did",
                "verkey",
                {"posted": True, "public": True},
                SOV,
                ED25519,
            )
        )
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.wallet_rotate_did_keypair(self.request)

    async def test_rotate_did_keypair_x(self):
        self.request.query = {"did": "did"}

        self.wallet.get_local_did = mock.CoroutineMock(
            return_value=DIDInfo(
                "did",
                "verkey",
                {"public": False},
                SOV,
                ED25519,
            )
        )
        self.wallet.rotate_did_keypair_start = mock.CoroutineMock(
            side_effect=test_module.WalletError()
        )
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.wallet_rotate_did_keypair(self.request)

    async def test_upgrade_anoncreds(self):
        self.profile.settings["wallet.name"] = "test_wallet"
        self.request.query = {"wallet_name": "not_test_wallet"}
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.upgrade_anoncreds(self.request)

        self.request.query = {"wallet_name": "not_test_wallet"}
        self.profile.settings["wallet.type"] = "askar-anoncreds"
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.upgrade_anoncreds(self.request)

        self.request.query = {"wallet_name": "test_wallet"}
        self.profile.settings["wallet.type"] = "askar"
        await test_module.upgrade_anoncreds(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



"""Schema artifacts."""

from marshmallow import fields

from ...messaging.models.openapi import OpenAPISchema
from ...messaging.valid import (
    INDY_SCHEMA_ID_EXAMPLE,
    INDY_SCHEMA_ID_VALIDATE,
    MAJOR_MINOR_VERSION_EXAMPLE,
    MAJOR_MINOR_VERSION_VALIDATE,
    NATURAL_NUM_EXAMPLE,
    NATURAL_NUM_VALIDATE,
)


class SchemaSchema(OpenAPISchema):
    """Marshmallow schema for indy schema."""

    ver = fields.Str(
        validate=MAJOR_MINOR_VERSION_VALIDATE,
        metadata={
            "description": "Node protocol version",
            "example": MAJOR_MINOR_VERSION_EXAMPLE,
        },
    )
    ident = fields.Str(
        data_key="id",
        validate=INDY_SCHEMA_ID_VALIDATE,
        metadata={
            "description": "Schema identifier",
            "example": INDY_SCHEMA_ID_EXAMPLE,
        },
    )
    name = fields.Str(
        metadata={
            "description": "Schema name",
            "example": INDY_SCHEMA_ID_EXAMPLE.split(":")[2],
        }
    )
    version = fields.Str(
        validate=MAJOR_MINOR_VERSION_VALIDATE,
        metadata={
            "description": "Schema version",
            "example": MAJOR_MINOR_VERSION_EXAMPLE,
        },
    )
    attr_names = fields.List(
        fields.Str(metadata={"description": "Attribute name", "example": "score"}),
        data_key="attrNames",
        metadata={"description": "Schema attribute names"},
    )
    seqNo = fields.Int(
        validate=NATURAL_NUM_VALIDATE,
        metadata={
            "description": "Schema sequence number",
            "strict": True,
            "example": NATURAL_NUM_EXAMPLE,
        },
    )



"""Message type identifiers for Trust Pings."""

from ...didcomm_prefix import DIDCommPrefix

SPEC_URI = (
    "https://github.com/decentralized-identity/aries-rfcs/tree/"
    "527849ec3aa2a8fd47a7bb6c57f918ff8bcb5e8c/features/0048-trust-ping"
)

# Message types
PING = "trust_ping/1.0/ping"
PING_RESPONSE = "trust_ping/1.0/ping_response"

PROTOCOL_PACKAGE = "acapy_agent.protocols.trustping.v1_0"

MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {
        PING: f"{PROTOCOL_PACKAGE}.messages.ping.Ping",
        PING_RESPONSE: f"{PROTOCOL_PACKAGE}.messages.ping_response.PingResponse",
    }
)



"""Credential proposal message handler."""

from .....indy.issuer import IndyIssuerError
from .....ledger.error import LedgerError
from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.models.base import BaseModelError
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageError
from .....utils.tracing import get_timer, trace_event
from .. import problem_report_for_record
from ..manager import CredentialManager, CredentialManagerError
from ..messages.credential_problem_report import ProblemReportReason
from ..messages.credential_proposal import CredentialProposal


class CredentialProposalHandler(BaseHandler):
    """Message handler class for credential proposals."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for credential proposals.

        Args:
            context: proposal context
            responder: responder callback

        """
        r_time = get_timer()
        profile = context.profile

        self._logger.debug("CredentialProposalHandler called with context %s", context)
        assert isinstance(context.message, CredentialProposal)
        self._logger.debug(
            "Received credential proposal message: %s",
            context.message.serialize(as_string=True),
        )

        # If connection is present it must be ready for use
        if context.connection_record and not context.connection_ready:
            raise HandlerException("Connection used for credential proposal not ready")
        elif not context.connection_record:
            raise HandlerException("Connectionless not supported for credential proposal")

        credential_manager = CredentialManager(profile)
        cred_ex_record = await credential_manager.receive_proposal(
            context.message, context.connection_record.connection_id
        )  # mgr only finds, saves record: on exception, saving state null is hopeless

        r_time = trace_event(
            context.settings,
            context.message,
            outcome="CredentialProposalHandler.handle.END",
            perf_counter=r_time,
        )

        # If auto_offer is enabled, respond immediately with offer
        if cred_ex_record.auto_offer:
            credential_offer_message = None
            try:
                (
                    cred_ex_record,
                    credential_offer_message,
                ) = await credential_manager.create_offer(
                    cred_ex_record,
                    counter_proposal=None,
                    comment=context.message.comment,
                )
                await responder.send_reply(credential_offer_message)
            except (
                BaseModelError,
                CredentialManagerError,
                IndyIssuerError,
                LedgerError,
                StorageError,
            ) as err:
                self._logger.exception("Error responding to credential proposal")
                if cred_ex_record:
                    async with profile.session() as session:
                        await cred_ex_record.save_error_state(
                            session,
                            reason=err.roll_up,  # us: be specific
                        )
                    await responder.send_reply(
                        problem_report_for_record(
                            cred_ex_record,
                            ProblemReportReason.ISSUANCE_ABANDONED.value,  # them: vague
                        )
                    )

            trace_event(
                context.settings,
                credential_offer_message,
                outcome="CredentialProposalHandler.handle.OFFER",
                perf_counter=r_time,
            )



from unittest import TestCase, mock

from ......messaging.decorators.attach_decorator import AttachDecorator
from .....didcomm_prefix import DIDCommPrefix
from ...message_types import ATTACH_DECO_IDS, CREDENTIAL_REQUEST, PROTOCOL_PACKAGE
from ..credential_request import CredentialRequest


class TestCredentialRequest(TestCase):
    """Credential request tests"""

    indy_cred_req = {
        "nonce": "1017762706737386703693758",
        "prover_did": "GMm4vMw8LLrLJjp81kRRLp",
        "cred_def_id": "GMm4vMw8LLrLJjp81kRRLp:3:CL:12:tag",
        "blinded_ms": {
            "u": "83907504917598709544715660183444547664806528194879236493704185267249518487609477830252206438464922282419526404954032744426656836343614241707982523911337758117991524606767981934822739259321023980818911648706424625657217291525111737996606024710795596961607334766957629765398381678917329471919374676824400143394472619220909211861028497009707890651887260349590274729523062264675018736459760546731362496666872299645586181905130659944070279943157241097916683504866583173110187429797028853314290183583689656212022982000994142291014801654456172923356395840313420880588404326139944888917762604275764474396403919497783080752861",
            "ur": "1 2422A7A25A9AB730F3399C77C28E1F6E02BB94A2C07D245B28DC4EE33E33DE49 1 1EF3FBD36FBA7510BDA79386508C0A84A33DF4171107C22895ACAE4FA4499F02 2 095E45DDF417D05FB10933FFC63D474548B7FFFF7888802F07FFFFFF7D07A8A8",
            "hidden_attributes": ["master_secret"],
            "committed_attributes": {},
        },
        "blinded_ms_correctness_proof": {
            "c": "77782990462020711078900471139684606615516190979556618670020830699801678914552",
            "v_dash_cap": "1966215015532422356590954855080129096516569112935438312989092847889400013191094311374123910677667707922694722167856889267996544544770134106600289624974901761453909338477897555013062690166110508298265469948048257876547569520215226798025984795668101468265482570744011744194025718081101032551943108999422057478928838218205736972438022128376728526831967897105301274481454020377656694232901381674223529320224276009919370080174601226836784570762698964476355045131401700464714725647784278935633253472872446202741297992383148244277451017022036452203286302631768247417186601621329239603862883753434562838622266122331169627284313213964584034951472090601638790603966977114416216909593408778336960753110805965734708636782885161632",
            "m_caps": {
                "master_secret": "1932933391026030434402535597188163725022560167138754201841873794167337347489231254032687761158191503499965986291267527620598858412377279828812688105949083285487853357240244045442"
            },
            "r_caps": {},
        },
    }

    cred_req = CredentialRequest(
        comment="Test",
        requests_attach=[
            AttachDecorator.data_base64(
                mapping=indy_cred_req,
                ident=ATTACH_DECO_IDS[CREDENTIAL_REQUEST],
            )
        ],
    )

    def test_init(self):
        """Test initializer"""
        credential_request = CredentialRequest(
            comment="Test",
            requests_attach=[
                AttachDecorator.data_base64(
                    mapping=self.indy_cred_req,
                    ident=ATTACH_DECO_IDS[CREDENTIAL_REQUEST],
                )
            ],
        )
        assert credential_request.requests_attach[0].content == self.indy_cred_req
        assert credential_request.indy_cred_req(0) == self.indy_cred_req

    def test_type(self):
        """Test type"""
        credential_request = CredentialRequest(
            comment="Test",
            requests_attach=[
                AttachDecorator.data_base64(
                    mapping=self.indy_cred_req,
                    ident=ATTACH_DECO_IDS[CREDENTIAL_REQUEST],
                )
            ],
        )

        assert credential_request._type == DIDCommPrefix.qualify_current(
            CREDENTIAL_REQUEST
        )

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.credential_request.CredentialRequestSchema.load"
    )
    def test_deserialize(self, mock_credential_request_schema_load):
        """
        Test deserialize
        """
        obj = self.indy_cred_req

        credential_request = CredentialRequest.deserialize(obj)
        mock_credential_request_schema_load.assert_called_once_with(obj)

        assert credential_request is mock_credential_request_schema_load.return_value

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.credential_request.CredentialRequestSchema.dump"
    )
    def test_serialize(self, mock_credential_request_schema_dump):
        """
        Test serialization.
        """
        credential_request = CredentialRequest(
            comment="Test",
            requests_attach=[
                AttachDecorator.data_base64(
                    mapping=self.indy_cred_req,
                    ident=ATTACH_DECO_IDS[CREDENTIAL_REQUEST],
                )
            ],
        )

        credential_request_dict = credential_request.serialize()
        mock_credential_request_schema_dump.assert_called_once_with(credential_request)

        assert credential_request_dict is mock_credential_request_schema_dump.return_value


class TestCredentialRequestSchema(TestCase):
    """Test credential cred request schema"""

    credential_request = CredentialRequest(
        comment="Test",
        requests_attach=[
            AttachDecorator.data_base64(TestCredentialRequest.indy_cred_req)
        ],
    )

    def test_make_model(self):
        """Test making model."""
        data = self.credential_request.serialize()
        model_instance = CredentialRequest.deserialize(data)
        assert isinstance(model_instance, CredentialRequest)



"""A presentation proposal content message."""

from typing import Optional, Sequence

from marshmallow import EXCLUDE, ValidationError, fields, validates_schema

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.decorators.attach_decorator import (
    AttachDecorator,
    AttachDecoratorSchema,
)
from ..message_types import PRES_20_PROPOSAL, PROTOCOL_PACKAGE
from .pres_format import V20PresFormat, V20PresFormatSchema

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers.pres_proposal_handler.V20PresProposalHandler"
)


class V20PresProposal(AgentMessage):
    """Class representing a presentation proposal."""

    class Meta:
        """V20PresProposal metadata."""

        handler_class = HANDLER_CLASS
        schema_class = "V20PresProposalSchema"
        message_type = PRES_20_PROPOSAL

    def __init__(
        self,
        _id: Optional[str] = None,
        *,
        comment: Optional[str] = None,
        formats: Sequence[V20PresFormat] = None,
        proposals_attach: Sequence[AttachDecorator] = None,
        **kwargs,
    ):
        """Initialize pres proposal object.

        Args:
            comment: optional human-readable comment
            formats: acceptable attachment formats
            proposals_attach: proposal attachments specifying criteria by format
            kwargs: additional key-value arguments
        """
        super().__init__(_id, **kwargs)
        self.comment = comment
        self.formats = list(formats) if formats else []
        self.proposals_attach = list(proposals_attach) if proposals_attach else []

    def attachment(self, fmt: V20PresFormat.Format = None) -> dict:
        """Return attached proposal item.

        Args:
            fmt: format of attachment in list to decode and return

        """
        target_format = (
            fmt
            if fmt
            else next(
                filter(
                    lambda ff: ff,
                    [V20PresFormat.Format.get(f.format) for f in self.formats],
                ),
                None,
            )
        )
        return (
            target_format.get_attachment_data(self.formats, self.proposals_attach)
            if target_format
            else None
        )


class V20PresProposalSchema(AgentMessageSchema):
    """Presentation proposal schema."""

    class Meta:
        """Presentation proposal schema metadata."""

        model_class = V20PresProposal
        unknown = EXCLUDE

    comment = fields.Str(
        required=False, metadata={"description": "Human-readable comment"}
    )
    formats = fields.Nested(
        V20PresFormatSchema,
        many=True,
        required=True,
        metadata={"description": "Acceptable attachment formats"},
    )
    proposals_attach = fields.Nested(
        AttachDecoratorSchema,
        many=True,
        required=True,
        data_key="proposals~attach",
        metadata={
            "description": (
                "Attachment per acceptable format on corresponding identifier"
            )
        },
    )

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate proposal attachment per format."""

        def get_attach_by_id(attach_id):
            """Return attachment with input attachment identifier."""
            for atch in attachments:
                if atch.ident == attach_id:
                    return atch
            raise ValidationError(f"No attachment for attach_id {attach_id} in formats")

        formats = data.get("formats") or []
        attachments = data.get("proposals_attach") or []
        if len(formats) != len(attachments):
            raise ValidationError("Formats/attachments length mismatch")

        for fmt in formats:
            atch = get_attach_by_id(fmt.attach_id)
            pres_format = V20PresFormat.Format.get(fmt.format)

            if pres_format:
                pres_format.validate_fields(PRES_20_PROPOSAL, atch.content)



from tempfile import NamedTemporaryFile
from unittest import IsolatedAsyncioTestCase

from ...cache.base import BaseCache
from ...core.plugin_registry import PluginRegistry
from ...core.profile import ProfileManager
from ...core.protocol_registry import ProtocolRegistry
from ...transport.wire_format import BaseWireFormat
from ..default_context import DefaultContextBuilder
from ..injection_context import InjectionContext


class TestDefaultContext(IsolatedAsyncioTestCase):
    async def test_build_context(self):
        """Test context init."""

        builder = DefaultContextBuilder()
        result = await builder.build_context()
        assert isinstance(result, InjectionContext)

        for cls in (
            BaseCache,
            BaseWireFormat,
            ProfileManager,
            ProtocolRegistry,
        ):
            assert isinstance(result.inject(cls), cls)

        builder = DefaultContextBuilder(
            settings={
                "timing.enabled": True,
                "timing.log.file": NamedTemporaryFile().name,
                "multitenant.enabled": True,
                "multitenant.admin_enabled": True,
            }
        )
        result = await builder.build_context()
        assert isinstance(result, InjectionContext)

    async def test_plugin_registration_askar_anoncreds(self):
        """Test anoncreds plugins are registered when wallet_type is askar-anoncreds."""
        builder = DefaultContextBuilder(
            settings={
                "wallet.type": "askar-anoncreds",
            }
        )
        result = await builder.build_context()
        plugin_registry = result.inject(PluginRegistry)

        # Check that anoncreds plugins are registered
        for plugin in [
            "acapy_agent.anoncreds",
            "acapy_agent.anoncreds.default.did_web",
            "acapy_agent.anoncreds.default.legacy_indy",
            "acapy_agent.revocation_anoncreds",
        ]:
            assert plugin in plugin_registry.plugin_names

    async def test_plugin_registration_multitenant_enabled(self):
        """Test anoncreds plugins are registered when multitenant is enabled."""
        builder = DefaultContextBuilder(
            settings={
                "multitenant.enabled": True,
            }
        )
        result = await builder.build_context()
        plugin_registry = result.inject(PluginRegistry)

        # Check that anoncreds and askar plugins are registered
        for plugin in [
            "acapy_agent.anoncreds",
            "acapy_agent.anoncreds.default.did_web",
            "acapy_agent.anoncreds.default.legacy_indy",
            "acapy_agent.revocation_anoncreds",
            "acapy_agent.messaging.credential_definitions",
            "acapy_agent.messaging.schemas",
            "acapy_agent.revocation",
        ]:
            assert plugin in plugin_registry.plugin_names

    async def test_plugin_registration_askar_only(self):
        """Test only askar plugins are registered when wallet_type is askar and multitenant is not enabled."""
        builder = DefaultContextBuilder(
            settings={
                "wallet.type": "askar",
                "multitenant.enabled": False,
            }
        )
        result = await builder.build_context()
        plugin_registry = result.inject(PluginRegistry)

        # Check that only askar plugins are registered
        for plugin in [
            "acapy_agent.messaging.credential_definitions",
            "acapy_agent.messaging.schemas",
            "acapy_agent.revocation",
        ]:
            assert plugin in plugin_registry.plugin_names

        # Ensure anoncreds plugins are not registered
        for plugin in [
            "acapy_agent.anoncreds",
            "acapy_agent.anoncreds.default.did_web",
            "acapy_agent.anoncreds.default.legacy_indy",
            "acapy_agent.revocation_anoncreds",
        ]:
            assert plugin not in plugin_registry.plugin_names



"""Base classes for Models and Schemas."""

import json
import logging
from abc import ABC
from collections import namedtuple
from typing import Mapping, Optional, Type, TypeVar, Union, cast, overload

from marshmallow import EXCLUDE, Schema, ValidationError, post_dump, post_load, pre_load
from typing_extensions import Literal

from ...core.error import BaseError
from ...utils.classloader import ClassLoader

LOGGER = logging.getLogger(__name__)

SerDe = namedtuple("SerDe", "ser de")


def resolve_class(the_cls, relative_cls: Optional[type] = None) -> type:
    """Resolve a class.

    Args:
        the_cls: The class to resolve
        relative_cls: Relative class to resolve from

    Returns:
        The resolved class

    Raises:
        ClassNotFoundError: If the class could not be loaded

    """
    resolved = None
    if isinstance(the_cls, type):
        resolved = the_cls
    elif isinstance(the_cls, str):
        default_module = relative_cls and relative_cls.__module__
        resolved = ClassLoader.load_class(the_cls, default_module)
    else:
        raise TypeError(
            f"Could not resolve class from {the_cls}; incorrect type {type(the_cls)}"
        )
    return resolved


def resolve_meta_property(obj, prop_name: str, defval=None):
    """Resolve a meta property.

    Args:
        obj: The object to resolve the property from
        prop_name: The property to resolve
        defval: The default value

    Returns:
        The meta property

    """
    if isinstance(obj, type):
        cls = obj
    else:
        cls = obj.__class__
    found = defval
    while cls:
        Meta = getattr(cls, "Meta", None)
        if Meta and hasattr(Meta, prop_name):
            found = getattr(Meta, prop_name)
            break
        cls = cls.__bases__[0]
        if cls is object:
            break
    return found


class BaseModelError(BaseError):
    """Base exception class for base model errors."""


ModelType = TypeVar("ModelType", bound="BaseModel")


class BaseModel(ABC):
    """Base model that provides convenience methods."""

    class Meta:
        """BaseModel meta data."""

        schema_class = None

    def __init__(self):
        """Initialize BaseModel.

        Raises:
            TypeError: If schema_class is not set on Meta

        """
        if not self.Meta.schema_class:
            raise TypeError(
                "Can't instantiate abstract class {} with no schema_class".format(
                    self.__class__.__name__
                )
            )

    @classmethod
    def _get_schema_class(cls) -> Type["BaseModelSchema"]:
        """Get the schema class.

        Returns:
            The resolved schema class

        """
        resolved = resolve_class(cls.Meta.schema_class, cls)
        if issubclass(resolved, BaseModelSchema):
            return resolved

        raise TypeError(
            f"Resolved class is not a subclass of BaseModelSchema: {resolved}"
        )

    @property
    def Schema(self) -> Type["BaseModelSchema"]:
        """Accessor for the model's schema class.

        Returns:
            The schema class

        """
        return self._get_schema_class()

    @overload
    @classmethod
    def deserialize(
        cls: Type[ModelType],
        obj,
    ) -> ModelType: ...

    @overload
    @classmethod
    def deserialize(
        cls: Type[ModelType],
        obj,
        *,
        unknown: Optional[str] = None,
    ) -> ModelType: ...

    @overload
    @classmethod
    def deserialize(
        cls: Type[ModelType],
        obj,
        *,
        none2none: Literal[False],
        unknown: Optional[str] = None,
    ) -> ModelType: ...

    @overload
    @classmethod
    def deserialize(
        cls: Type[ModelType],
        obj,
        *,
        none2none: Literal[True],
        unknown: Optional[str] = None,
    ) -> Optional[ModelType]: ...

    @classmethod
    def deserialize(
        cls: Type[ModelType],
        obj,
        *,
        unknown: Optional[str] = None,
        none2none: bool = False,
    ) -> Optional[ModelType]:
        """Convert from JSON representation to a model instance.

        Args:
            obj: The dict to load into a model instance
            unknown: Behaviour for unknown attributes
            none2none: Deserialize None to None

        Returns:
            A model instance for this data

        """
        if obj is None and none2none:
            return None

        schema_cls = cls._get_schema_class()
        schema = schema_cls(
            unknown=unknown or resolve_meta_property(schema_cls, "unknown", EXCLUDE)
        )

        try:
            return cast(
                ModelType,
                schema.loads(obj) if isinstance(obj, str) else schema.load(obj),
            )
        except (AttributeError, ValidationError) as err:
            LOGGER.exception(f"{cls.__name__} message validation error:")
            raise BaseModelError(f"{cls.__name__} schema validation failed") from err

    @overload
    def serialize(
        self,
        *,
        as_string: Literal[True],
        unknown: Optional[str] = None,
    ) -> str: ...

    @overload
    def serialize(
        self,
        *,
        unknown: Optional[str] = None,
    ) -> dict: ...

    def serialize(
        self,
        *,
        as_string: bool = False,
        unknown: Optional[str] = None,
    ) -> Union[str, dict]:
        """Create a JSON-compatible dict representation of the model instance.

        Args:
            as_string: Return a string of JSON instead of a dict
            unknown: Behaviour for unknown attributes

        Returns:
            A dict representation of this model, or a JSON string if as_string is True

        """
        schema_cls = self._get_schema_class()
        schema = schema_cls(
            unknown=unknown or resolve_meta_property(schema_cls, "unknown", EXCLUDE)
        )
        try:
            return (
                schema.dumps(self, separators=(",", ":"))
                if as_string
                else schema.dump(self)
            )
        except (AttributeError, ValidationError, TypeError) as err:
            LOGGER.exception(f"{self.__class__.__name__} message serialization error:")
            raise BaseModelError(
                f"{self.__class__.__name__} schema validation failed"
            ) from err

    @classmethod
    def serde(cls, obj: Union["BaseModel", Mapping, None]) -> Optional[SerDe]:
        """Return serialized, deserialized representations of input object."""
        if obj is None:
            return None

        if isinstance(obj, BaseModel):
            return SerDe(obj.serialize(), obj)

        return SerDe(obj, cls.deserialize(obj))

    def validate(self, unknown: Optional[str] = None):
        """Validate a constructed model."""
        schema = self.Schema(unknown=unknown)
        errors = schema.validate(self.serialize())
        if errors:
            raise ValidationError(errors)
        return self

    @classmethod
    def from_json(
        cls,
        json_repr: Union[str, bytes],
        unknown: Optional[str] = None,
    ):
        """Parse a JSON string into a model instance.

        Args:
            json_repr: JSON string
            unknown: Behaviour for unknown attributes

        Returns:
            A model instance representation of this JSON

        """
        try:
            parsed = json.loads(json_repr)
        except ValueError as e:
            LOGGER.exception(f"{cls.__name__} message parse error:")
            raise BaseModelError(f"{cls.__name__} JSON parsing failed") from e
        return cls.deserialize(parsed, unknown=unknown)

    def to_json(self, unknown: Optional[str] = None) -> str:
        """Create a JSON representation of the model instance.

        Returns:
            A JSON representation of this message

        """
        return json.dumps(self.serialize(unknown=unknown))

    def __repr__(self) -> str:
        """Return a human readable representation of this class.

        Returns:
            A human readable string for this class

        """
        exclude = resolve_meta_property(self, "repr_exclude", [])
        items = (
            "{}={}".format(k, repr(v))
            for k, v in self.__dict__.items()
            if k not in exclude
        )
        return "<{}({})>".format(self.__class__.__name__, ", ".join(items))


class BaseModelSchema(Schema):
    """BaseModel schema."""

    class Meta:
        """BaseModelSchema metadata."""

        model_class = None
        skip_values = [None]

    def __init__(self, *args, **kwargs):
        """Initialize BaseModelSchema.

        Raises:
            TypeError: If model_class is not set on Meta

        """
        super().__init__(*args, **kwargs)
        if not self.Meta.model_class:
            raise TypeError(
                "Can't instantiate abstract class {} with no model_class".format(
                    self.__class__.__name__
                )
            )

    @classmethod
    def _get_model_class(cls):
        """Get the model class.

        Returns:
            The model class

        """
        return resolve_class(cls.Meta.model_class, cls)

    @property
    def Model(self) -> type:
        """Accessor for the schema's model class.

        Returns:
            The model class

        """
        return self._get_model_class()

    @pre_load
    def skip_dump_only(self, data, **kwargs):
        """Skip fields that are only expected during serialization.

        Args:
            data: The incoming data to clean
            kwargs: Arbitrary keyword arguments

        Returns:
            The modified data

        """
        if not data:
            return data

        to_remove = {
            field_obj.data_key or field_name
            for field_name, field_obj in self.fields.items()
            if field_obj.dump_only
        }
        for field_name in to_remove:
            if field_name in data:
                del data[field_name]
        return data

    @post_load
    def make_model(self, data: dict, **kwargs):
        """Return model instance after loading.

        Returns:
            A model instance

        """
        try:
            cls_inst = self.Model(**data)
        except TypeError as err:
            if "_type" in str(err) and "_type" in data:
                data["msg_type"] = data["_type"]
                del data["_type"]
            cls_inst = self.Model(**data)
        return cls_inst

    @post_dump
    def remove_skipped_values(self, data, **kwargs):
        """Remove values that are are marked to skip.

        Returns:
            Returns this modified data

        """
        skip_vals = resolve_meta_property(self, "skip_values", [])
        return {key: value for key, value in data.items() if value not in skip_vals}



from unittest import IsolatedAsyncioTestCase

from ......indy.models.pres_preview import (
    IndyPresAttrSpec,
    IndyPresPredSpec,
    IndyPresPreview,
)
from ......messaging.models.base_record import BaseExchangeRecord, BaseExchangeSchema
from ......tests import mock
from ......utils.testing import create_test_profile
from ...messages.presentation_proposal import PresentationProposal
from .. import presentation_exchange as test_module
from ..presentation_exchange import V10PresentationExchange

S_ID = "NcYxiDXkpYi6ov5FcYDi1e:2:vidya:1.0"
CD_ID = f"NcYxiDXkpYi6ov5FcYDi1e:3:CL:{S_ID}:tag1"
INDY_PROOF_REQ = {
    "name": "proof-req",
    "version": "1.0",
    "nonce": "12345",
    "requested_attributes": {
        "0_player_uuid": {
            "name": "player",
            "restrictions": [
                {
                    "cred_def_id": f"{CD_ID}",
                    "attr::player::value": "Richie Knucklez",
                }
            ],
            "non_revoked": {
                "from": 1234567890,
                "to": 1234567890,
            },
        },
        "0_screencapture_uuid": {
            "name": "screenCapture",
            "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            "non_revoked": {
                "from": 1234567890,
                "to": 1234567890,
            },
        },
    },
    "requested_predicates": {
        "0_highscore_GE_uuid": {
            "name": "highScore",
            "p_type": ">=",
            "p_value": 1000000,
            "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            "non_revoked": {
                "from": 1234567890,
                "to": 1234567890,
            },
        }
    },
}
PRES_PREVIEW = IndyPresPreview(
    attributes=[
        IndyPresAttrSpec(name="player", cred_def_id=CD_ID, value="Richie Knucklez"),
        IndyPresAttrSpec(
            name="screenCapture",
            cred_def_id=CD_ID,
            mime_type="image/png",
            value="aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
        ),
    ],
    predicates=[
        IndyPresPredSpec(
            name="highScore", cred_def_id=CD_ID, predicate=">=", threshold=1000000
        )
    ],
)


class BasexRecordImpl(BaseExchangeRecord):
    class Meta:
        schema_class = "BasexRecordImplSchema"

    RECORD_TYPE = "record"


class BasexRecordImplSchema(BaseExchangeSchema):
    class Meta:
        model_class = BasexRecordImpl


class TestRecord(IsolatedAsyncioTestCase):
    async def test_record(self):
        presentation_proposal = PresentationProposal(
            comment="Hello World", presentation_proposal=PRES_PREVIEW
        )
        record = V10PresentationExchange(
            presentation_exchange_id="pxid",
            connection_id="conn_id",
            thread_id="thid",
            auto_present=True,
            auto_remove=True,
        )
        record.presentation_proposal_dict = presentation_proposal  # cover setter
        record.presentation_request_dict = None  # cover setter

        assert record.presentation_exchange_id == "pxid"

        assert record.record_value == {
            "connection_id": "conn_id",
            "initiator": None,
            "presentation_proposal_dict": presentation_proposal.serialize(),
            "role": None,
            "state": None,
            "auto_present": True,
            "auto_verify": False,
            "error_msg": None,
            "verified": None,
            "verified_msgs": None,
            "trace": False,
            "auto_remove": True,
        }

        bx_record = BasexRecordImpl()
        assert record != bx_record

    async def test_save_error_state(self):
        self.profile = await create_test_profile()
        record = V10PresentationExchange(state=None)
        assert record._last_state is None
        async with self.profile.session() as session:
            await record.save_error_state(session)  # cover short circuit

            record.state = V10PresentationExchange.STATE_PROPOSAL_RECEIVED
            await record.save(session)

            with (
                mock.patch.object(record, "save", mock.CoroutineMock()) as mock_save,
                mock.patch.object(
                    test_module.LOGGER, "exception", mock.MagicMock()
                ) as mock_log_exc,
            ):
                mock_save.side_effect = test_module.StorageError()
                await record.save_error_state(session, reason="testing")
                mock_log_exc.assert_called_once()



"""Timed rotating file handler for aca-py agent."""

import logging
import os
import re
import time as mod_time
from datetime import datetime, timedelta
from logging.handlers import BaseRotatingHandler
from random import randint

from portalocker import LOCK_EX, lock, unlock


class TimedRotatingFileMultiProcessHandler(BaseRotatingHandler):
    """Handler for logging to a file with timed rotation and multi-process support.

    Derived from Python's `logging.handlers` and custom implementations to handle
    multi-process scenarios.

    This implementation is based on Python's built-in logging handlers and the mpfhandler
    library, but includes custom modifications to properly handle interval, backupCount,
    and deletion of old backup files during rotation in multi-process scenarios.

    References:
        - https://github.com/python/cpython/blob/main/Lib/logging/handlers.py
        - https://github.com/yorks/mpfhandler/blob/master/src/mpfhandler.py
    """

    def __init__(
        self,
        filename,
        when="h",
        interval=1,
        backupCount=1,
        encoding=None,
        delay=False,
        utc=False,
        atTime=None,
    ):
        """Initialize an instance of `TimedRotatingFileMultiProcessHandler`.

        Args:
            filename (str): The log file name with path.
            when (str, optional): Specifies when to rotate the log file. Defaults to "h".
            interval (int, optional): The interval when to rotate. Defaults to 1.
            backupCount (int, optional): The count of backup files. A backupCount of 0
                means no limit on the count of backup files (no backup will be deleted).
                Defaults to 1.
            encoding (str, optional): The encoding to use for the log file.
                Defaults to None.
            delay (bool, optional): Whether to delay file opening until the first log
                message is emitted. Defaults to False.
            utc (bool, optional): Whether to use UTC time for log rotation.
                Defaults to False.
            atTime (datetime.time, optional): The specific time at which log rotation
                should occur. Defaults to None.

        Raises:
            ValueError: If an invalid rollover interval is specified.

        """
        BaseRotatingHandler.__init__(
            self,
            filename,
            "a",
            encoding=encoding,
            delay=delay,
        )
        self.when = when.upper()
        self.backupCount = backupCount
        self.utc = utc
        self.atTime = atTime
        self.mylogfile = "%s.%08d" % ("/tmp/trfmphanldler", randint(0, 99999999))
        self.interval = interval

        # Set the appropriate suffix and regular expression pattern based on the specified rollover interval # noqa: E501
        if self.when == "S":
            self.interval = 1
            self.suffix = "%Y-%m-%d_%H-%M-%S"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}(\.\w+)?$"
        elif self.when == "M":
            self.interval = 60
            self.suffix = "%Y-%m-%d_%H-%M"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}-\d{2}(\.\w+)?$"
        elif self.when == "H":
            self.interval = 60 * 60
            self.suffix = "%Y-%m-%d_%H"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}(\.\w+)?$"
        elif self.when == "D" or self.when == "MIDNIGHT":
            self.interval = 60 * 60 * 24
            self.suffix = "%Y-%m-%d"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}(\.\w+)?$"
        elif self.when.startswith("W"):
            self.interval = 60 * 60 * 24 * 7
            if len(self.when) != 2:
                raise ValueError(
                    "You must specify a day for weekly rollover from 0 "
                    "to 6 (0 is Monday): %s" % self.when
                )
            if self.when[1] < "0" or self.when[1] > "6":
                raise ValueError(
                    "Invalid day specified for weekly rollover: %s" % self.when
                )
            self.dayOfWeek = int(self.when[1])
            self.suffix = "%Y-%m-%d"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}(\.\w+)?$"
        else:
            raise ValueError("Invalid rollover interval specified: %s" % self.when)

        self.extMatch = re.compile(self.extMatch, re.ASCII)
        self.interval = self.interval * interval
        self.stream_lock = None
        self.lock_file = self._getLockFile()
        self.next_rollover_time = self.get_next_rollover_time()
        if not self.next_rollover_time:
            self.next_rollover_time = self.compute_next_rollover_time()
            self.save_next_rollover_time()

    def _log2mylog(self, msg):
        """Write to external log file."""
        time_str = mod_time.strftime(
            "%Y-%m-%d %H:%M:%S", mod_time.localtime(mod_time.time())
        )
        msg = str(msg)
        content = "%s [%s]\n" % (time_str, msg)
        fa = open(self.mylogfile, "a")
        fa.write(content)
        fa.close()

    def _getLockFile(self):
        """Return log lock file."""
        if self.baseFilename.endswith(".log"):
            lock_file = self.baseFilename[:-4]
        else:
            lock_file = self.baseFilename
        lock_file += ".lock"
        return lock_file

    def _openLockFile(self):
        """Open log lock file."""
        lock_file = self._getLockFile()
        self.stream_lock = open(lock_file, "w")

    def compute_next_rollover_time(self):
        """Return next rollover time."""
        next_time = None
        current_datetime = datetime.now()
        if self.when == "D":
            next_datetime = current_datetime + timedelta(days=self.interval)
            next_date = next_datetime.date()
            next_time = int(mod_time.mktime(next_date.timetuple()))
        elif self.when.startswith("W"):
            days = 0
            current_weekday = current_datetime.weekday()
            if current_weekday == self.dayOfWeek:
                days = self.interval + 7
            elif current_weekday < self.dayOfWeek:
                days = self.dayOfWeek - current_weekday
            else:
                days = 6 - current_weekday + self.dayOfWeek + 1
            next_datetime = current_datetime + timedelta(days=days)
            next_date = next_datetime.date()
            next_time = int(mod_time.mktime(next_date.timetuple()))
        else:
            tmp_next_datetime = current_datetime + timedelta(seconds=self.interval)
            next_datetime = tmp_next_datetime.replace(microsecond=0)
            if self.when == "H":
                next_datetime = tmp_next_datetime.replace(
                    minute=0, second=0, microsecond=0
                )
            elif self.when == "M":
                next_datetime = tmp_next_datetime.replace(second=0, microsecond=0)
            next_time = int(mod_time.mktime(next_datetime.timetuple()))
        return next_time

    def get_next_rollover_time(self):
        """Get next rollover time stamp from lock file."""
        try:
            fp = open(self.lock_file, "r")
            c = fp.read()
            fp.close()
            return int(c)
        except Exception:
            return False

    def save_next_rollover_time(self):
        """Save the nextRolloverTimestamp to lock file."""
        if not self.next_rollover_time:
            return 0
        content = "%d" % self.next_rollover_time
        if not self.stream_lock:
            self._openLockFile()
        lock(self.stream_lock, LOCK_EX)
        try:
            self.stream_lock.seek(0)
            self.stream_lock.write(content)
            self.stream_lock.flush()
        except Exception:
            pass
        finally:
            unlock(self.stream_lock)

    def acquire(self):
        """Acquire thread and file locks."""
        BaseRotatingHandler.acquire(self)
        if self.stream_lock:
            if self.stream_lock.closed:
                try:
                    self._openLockFile()
                except Exception:
                    self.stream_lock = None
                    return
            lock(self.stream_lock, LOCK_EX)

    def release(self):
        """Release file and thread locks."""
        try:
            if self.stream_lock and not self.stream_lock.closed:
                unlock(self.stream_lock)
        except Exception:
            pass
        finally:
            BaseRotatingHandler.release(self)

    def _close_stream(self):
        """Close the log file stream."""
        if self.stream:
            try:
                if not self.stream.closed:
                    self.stream.flush()
                    self.stream.close()
            finally:
                self.stream = None

    def _close_stream_lock(self):
        """Close the lock file stream."""
        if self.stream_lock:
            try:
                if not self.stream_lock.closed:
                    self.stream_lock.flush()
                    self.stream_lock.close()
            finally:
                self.stream_lock = None

    def close(self):
        """Close log stream and stream_lock."""
        try:
            self._close_stream()
            self._close_stream_lock()
        finally:
            self.stream = None
            self.stream_lock = None

    def get_log_files_to_delete(self):
        """Delete backup files on rotation based on backupCount."""
        dir_name, base_name = os.path.split(self.baseFilename)
        file_names = os.listdir(dir_name)
        result = []
        n, e = os.path.splitext(base_name)
        prefix = n + "."
        plen = len(prefix)
        for file_name in file_names:
            if self.namer is None:
                if not file_name.startswith(base_name):
                    continue
            else:
                if (
                    not file_name.startswith(base_name)
                    and file_name.endswith(e)
                    and len(file_name) > (plen + 1)
                    and not file_name[plen + 1].isdigit()
                ):
                    continue
            if file_name[:plen] == prefix:
                suffix = file_name[plen:]
                parts = suffix.split(".")
                for part in parts:
                    if self.extMatch.match(part):
                        result.append(os.path.join(dir_name, file_name))
                        break
        if len(result) < self.backupCount:
            result = []
        else:
            result.sort()
            result = result[: len(result) - self.backupCount]
        return result

    def shouldRollover(self, record):
        """Determine if rollover should occur."""
        t = int(mod_time.time())
        if t >= self.next_rollover_time:
            return 1
        return 0

    def doRollover(self):
        """Perform rollover."""
        self._close_stream()
        self.acquire()
        try:
            file_next_rollover_time = self.get_next_rollover_time()
            if not file_next_rollover_time:
                self.release()
                return 0
            if self.next_rollover_time < file_next_rollover_time:
                self.next_rollover_time = file_next_rollover_time
                self.release()
                return 0
        except Exception:
            pass
        time_tuple = mod_time.localtime(self.next_rollover_time - 1)
        dfn = self.baseFilename + "." + mod_time.strftime(self.suffix, time_tuple)
        if os.path.exists(dfn):
            bakname = dfn + ".bak"
            while os.path.exists(bakname):
                bakname = "%s.%08d" % (bakname, randint(0, 99999999))
            try:
                os.rename(dfn, bakname)
            except Exception:
                pass
        if os.path.exists(self.baseFilename):
            try:
                os.rename(self.baseFilename, dfn)
            except Exception:
                pass
        self.next_rollover_time = self.compute_next_rollover_time()
        self.save_next_rollover_time()
        if self.backupCount > 0:
            for s in self.get_log_files_to_delete():
                os.remove(s)
        if not self.delay:
            self.stream = self._open()
        self.release()


logging.handlers.TimedRotatingFileMultiProcessHandler = (
    TimedRotatingFileMultiProcessHandler
)



"""A credential preview inner object."""

from typing import Optional, Sequence

from marshmallow import EXCLUDE, fields

from ......messaging.models.base import BaseModel, BaseModelSchema
from ......wallet.util import b64_to_str
from .....didcomm_prefix import DIDCommPrefix
from ...message_types import CREDENTIAL_PREVIEW


class CredAttrSpec(BaseModel):
    """Class representing a preview of an attibute."""

    class Meta:
        """Attribute preview metadata."""

        schema_class = "CredAttrSpecSchema"

    def __init__(
        self, *, name: str, value: str, mime_type: Optional[str] = None, **kwargs
    ):
        """Initialize attribute preview object.

        Args:
            name: attribute name
            value: attribute value; caller must base64-encode for attributes with
                non-empty MIME type
            mime_type: MIME type (default null)
            kwargs: additional keyword arguments to map into message class properties

        """
        super().__init__(**kwargs)

        self.name = name
        self.value = value
        self.mime_type = mime_type.lower() if mime_type else None

    @staticmethod
    def list_plain(plain: dict):
        """Return a list of `CredAttrSpec` without MIME types from names/values.

        Args:
            plain: dict mapping names to values

        Returns:
            List of CredAttrSpecs with no MIME types

        """
        return [CredAttrSpec(name=k, value=plain[k]) for k in plain]

    def b64_decoded_value(self) -> str:
        """Value, base64-decoded if applicable."""

        return b64_to_str(self.value) if self.value and self.mime_type else self.value

    def __eq__(self, other):
        """Equality comparator."""

        if self.name != other.name:
            return False  # distinct attribute names

        if self.mime_type != other.mime_type:
            return False  # distinct MIME types

        return self.b64_decoded_value() == other.b64_decoded_value()


class CredAttrSpecSchema(BaseModelSchema):
    """Attribute preview schema."""

    class Meta:
        """Attribute preview schema metadata."""

        model_class = CredAttrSpec
        unknown = EXCLUDE

    name = fields.Str(
        required=True,
        metadata={"description": "Attribute name", "example": "favourite_drink"},
    )
    mime_type = fields.Str(
        required=False,
        data_key="mime-type",
        allow_none=True,
        metadata={
            "description": "MIME type: omit for (null) default",
            "example": "image/jpeg",
        },
    )
    value = fields.Str(
        required=True,
        metadata={
            "description": "Attribute value: base64-encode if MIME type is present",
            "example": "martini",
        },
    )


class CredentialPreview(BaseModel):
    """Class representing a credential preview inner object."""

    class Meta:
        """Credential preview metadata."""

        schema_class = "CredentialPreviewSchema"
        message_type = CREDENTIAL_PREVIEW

    def __init__(
        self,
        *,
        _type: Optional[str] = None,
        attributes: Sequence[CredAttrSpec] = None,
        **kwargs,
    ):
        """Initialize credential preview object.

        Args:
            _type: formalism for Marshmallow model creation: ignored
            attributes (list): list of attribute preview dicts; e.g., [
                {
                    "name": "attribute_name",
                    "value": "value"
                },
                {
                    "name": "icon",
                    "mime-type": "image/png",
                    "value": "cG90YXRv"
                }
            ]
            kwargs: additional keyword arguments to map into message class properties

        """
        super().__init__(**kwargs)
        self.attributes = list(attributes) if attributes else []

    @property
    def _type(self):
        """Accessor for message type."""
        return DIDCommPrefix.qualify_current(CredentialPreview.Meta.message_type)

    def attr_dict(self, decode: bool = False):
        """Return name:value pair per attribute.

        Args:
            decode: whether first to decode attributes with MIME type

        """

        return {
            attr.name: (
                b64_to_str(attr.value) if attr.mime_type and decode else attr.value
            )
            for attr in self.attributes
        }

    def mime_types(self):
        """Return per-attribute mapping from name to MIME type.

        Return empty dict if no attribute has MIME type.

        """
        return {attr.name: attr.mime_type for attr in self.attributes if attr.mime_type}


class CredentialPreviewSchema(BaseModelSchema):
    """Credential preview schema."""

    class Meta:
        """Credential preview schema metadata."""

        model_class = CredentialPreview
        unknown = EXCLUDE

    _type = fields.Str(
        required=False,
        data_key="@type",
        metadata={
            "description": "Message type identifier",
            "example": CREDENTIAL_PREVIEW,
        },
    )
    attributes = fields.Nested(
        CredAttrSpecSchema, many=True, required=True, data_key="attributes"
    )



"""Store revocation notification details until revocation is published."""

from typing import Optional, Sequence

from marshmallow import fields
from marshmallow.utils import EXCLUDE

from .....core.profile import ProfileSession
from .....messaging.models.base_record import BaseRecord, BaseRecordSchema
from .....messaging.valid import (
    INDY_CRED_REV_ID_EXAMPLE,
    INDY_CRED_REV_ID_VALIDATE,
    INDY_REV_REG_ID_EXAMPLE,
    INDY_REV_REG_ID_VALIDATE,
    UUID4_EXAMPLE,
    UUID4_VALIDATE,
)
from .....storage.error import StorageDuplicateError, StorageNotFoundError
from ..messages.revoke import Revoke


class RevNotificationRecord(BaseRecord):
    """Revocation Notification Record."""

    class Meta:
        """RevNotificationRecord Meta."""

        schema_class = "RevNotificationRecordSchema"

    RECORD_TYPE = "revocation_notification"
    RECORD_ID_NAME = "revocation_notification_id"
    TAG_NAMES = {
        "rev_reg_id",
        "cred_rev_id",
        "connection_id",
        "version",
    }

    def __init__(
        self,
        *,
        revocation_notification_id: Optional[str] = None,
        rev_reg_id: Optional[str] = None,
        cred_rev_id: Optional[str] = None,
        connection_id: Optional[str] = None,
        thread_id: Optional[str] = None,
        comment: Optional[str] = None,
        version: Optional[str] = None,
        **kwargs,
    ):
        """Construct record."""
        super().__init__(revocation_notification_id, **kwargs)
        self.rev_reg_id = rev_reg_id
        self.cred_rev_id = cred_rev_id
        self.connection_id = connection_id
        self.thread_id = thread_id
        self.comment = comment
        self.version = version

    @property
    def revocation_notification_id(self) -> Optional[str]:
        """Return record id."""
        return self._id

    @property
    def record_value(self) -> dict:
        """Return record value."""
        return {prop: getattr(self, prop) for prop in ("thread_id", "comment")}

    @classmethod
    async def query_by_ids(
        cls,
        session: ProfileSession,
        cred_rev_id: str,
        rev_reg_id: str,
    ) -> "RevNotificationRecord":
        """Retrieve revocation notification record by cred rev id and/or rev reg id.

        Args:
            session: the profile session to use
            cred_rev_id: the cred rev id by which to filter
            rev_reg_id: the rev reg id by which to filter
        """
        tag_filter = {
            **{"version": "v2_0"},
            **{"cred_rev_id": cred_rev_id for _ in [""] if cred_rev_id},
            **{"rev_reg_id": rev_reg_id for _ in [""] if rev_reg_id},
        }

        result = await cls.query(session, tag_filter)
        if len(result) > 1:
            raise StorageDuplicateError(
                "More than one RevNotificationRecord was found for the given IDs"
            )
        if not result:
            raise StorageNotFoundError("No RevNotificationRecord found for the given IDs")
        return result[0]

    @classmethod
    async def query_by_rev_reg_id(
        cls,
        session: ProfileSession,
        rev_reg_id: str,
    ) -> Sequence["RevNotificationRecord"]:
        """Retrieve revocation notification records by rev reg id.

        Args:
            session: the profile session to use
            rev_reg_id: the rev reg id by which to filter
        """
        tag_filter = {
            **{"version": "v2_0"},
            **{"rev_reg_id": rev_reg_id for _ in [""] if rev_reg_id},
        }

        return await cls.query(session, tag_filter)

    def to_message(self):
        """Return a revocation notification constructed from this record."""
        if not self.thread_id:
            raise ValueError(
                "No thread ID set on revocation notification record, "
                "cannot create message"
            )
        return Revoke(
            revocation_format="indy-anoncreds",
            credential_id=f"{self.rev_reg_id}::{self.cred_rev_id}",
            comment=self.comment,
        )


class RevNotificationRecordSchema(BaseRecordSchema):
    """Revocation Notification Record Schema."""

    class Meta:
        """RevNotificationRecordSchema Meta."""

        model_class = "RevNotificationRecord"
        unknown = EXCLUDE

    rev_reg_id = fields.Str(
        required=False,
        validate=INDY_REV_REG_ID_VALIDATE,
        metadata={
            "description": "Revocation registry identifier",
            "example": INDY_REV_REG_ID_EXAMPLE,
        },
    )
    cred_rev_id = fields.Str(
        required=False,
        validate=INDY_CRED_REV_ID_VALIDATE,
        metadata={
            "description": "Credential revocation identifier",
            "example": INDY_CRED_REV_ID_EXAMPLE,
        },
    )
    connection_id = fields.Str(
        required=False,
        validate=UUID4_VALIDATE,
        metadata={
            "description": (
                "Connection ID to which the revocation notification will be sent;"
                " required if notify is true"
            ),
            "example": UUID4_EXAMPLE,
        },
    )
    thread_id = fields.Str(
        required=False,
        metadata={
            "description": (
                "Thread ID of the credential exchange message thread resulting in the"
                " credential now being revoked; required if notify is true"
            )
        },
    )
    comment = fields.Str(
        required=False,
        metadata={
            "description": "Optional comment to include in revocation notification"
        },
    )
    version = fields.Str(
        required=False,
        metadata={"description": "Version of Revocation Notification to send out"},
    )



from datetime import datetime, timedelta
from unittest import IsolatedAsyncioTestCase, mock

from ....tests.data import TEST_VC_DOCUMENT_SIGNED_DID_KEY_ED25519
from ....tests.document_loader import custom_document_loader
from ..controller_proof_purpose import ControllerProofPurpose
from ..proof_purpose import ProofPurpose


class TestControllerProofPurpose(IsolatedAsyncioTestCase):
    async def test_properties(self):
        term = "TestTerm"
        date = datetime.now()
        delta = timedelta(1)
        proof_purpose = ControllerProofPurpose(
            term=term, date=date, max_timestamp_delta=delta
        )
        proof_purpose2 = ControllerProofPurpose(
            term=term, date=date, max_timestamp_delta=delta
        )

        assert proof_purpose.term == term
        assert proof_purpose.date == date
        assert proof_purpose.max_timestamp_delta == delta

        assert proof_purpose2 == proof_purpose
        assert proof_purpose != 10

    async def test_validate(self):
        proof_purpose = ControllerProofPurpose(term="assertionMethod")

        document = TEST_VC_DOCUMENT_SIGNED_DID_KEY_ED25519.copy()
        proof = document.pop("proof")
        suite = mock.MagicMock()
        verification_method = {
            "id": TEST_VC_DOCUMENT_SIGNED_DID_KEY_ED25519["proof"]["verificationMethod"],
            "controller": TEST_VC_DOCUMENT_SIGNED_DID_KEY_ED25519["issuer"],
        }

        result = proof_purpose.validate(
            proof=proof,
            document=document,
            suite=suite,
            verification_method=verification_method,
            document_loader=custom_document_loader,
        )

        assert result.valid

    async def test_validate_controller_invalid_type(self):
        proof_purpose = ControllerProofPurpose(term="assertionMethod")

        document = TEST_VC_DOCUMENT_SIGNED_DID_KEY_ED25519.copy()
        proof = document.pop("proof")
        suite = mock.MagicMock()
        verification_method = {
            "id": TEST_VC_DOCUMENT_SIGNED_DID_KEY_ED25519["proof"]["verificationMethod"],
            "controller": 10,
        }

        result = proof_purpose.validate(
            proof=proof,
            document=document,
            suite=suite,
            verification_method=verification_method,
            document_loader=custom_document_loader,
        )

        assert not result.valid
        assert '"controller" must be a string or dict' in str(result.error)

    async def test_validate_x_not_authorized(self):
        proof_purpose = ControllerProofPurpose(term="assertionMethod")

        document = TEST_VC_DOCUMENT_SIGNED_DID_KEY_ED25519.copy()
        proof = document.pop("proof")
        suite = mock.MagicMock()
        verification_method = {
            "id": TEST_VC_DOCUMENT_SIGNED_DID_KEY_ED25519["proof"]["verificationMethod"],
            "controller": "did:example:489398593",
        }

        result = proof_purpose.validate(
            proof=proof,
            document=document,
            suite=suite,
            verification_method=verification_method,
            document_loader=custom_document_loader,
        )

        assert not result.valid
        assert "not authorized by controller for proof purpose" in str(result.error)

    async def test_validate_x_super_invalid(self):
        proof_purpose = ControllerProofPurpose(term="assertionMethod")

        with mock.patch.object(ProofPurpose, "validate") as validate_mock:
            validate_mock.return_value = mock.MagicMock(valid=False)

            result = proof_purpose.validate(
                proof=mock.MagicMock(),
                document=mock.MagicMock(),
                suite=mock.MagicMock(),
                verification_method=mock.MagicMock(),
                document_loader=mock.MagicMock(),
            )

            assert not result.valid
            assert result == validate_mock.return_value



"""Action menu admin routes."""

import logging

from aiohttp import web
from aiohttp_apispec import docs, match_info_schema, request_schema, response_schema
from marshmallow import fields

from ....admin.decorators.auth import tenant_authentication
from ....admin.request_context import AdminRequestContext
from ....connections.models.conn_record import ConnRecord
from ....messaging.models.base import BaseModelError
from ....messaging.models.openapi import OpenAPISchema
from ....messaging.valid import UUID4_EXAMPLE
from ....storage.error import StorageError, StorageNotFoundError
from .messages.menu import Menu, MenuSchema
from .messages.menu_request import MenuRequest
from .messages.perform import Perform
from .models.menu_option import MenuOptionSchema
from .util import MENU_RECORD_TYPE, retrieve_connection_menu, save_connection_menu

LOGGER = logging.getLogger(__name__)


class ActionMenuModulesResultSchema(OpenAPISchema):
    """Schema for the modules endpoint."""


class PerformRequestSchema(OpenAPISchema):
    """Request schema for performing a menu action."""

    name = fields.Str(metadata={"description": "Menu option name", "example": "Query"})
    params = fields.Dict(
        required=False,
        keys=fields.Str(metadata={"example": "parameter"}),
        values=fields.Str(metadata={"example": UUID4_EXAMPLE}),
        metadata={"description": "Input parameter values"},
    )


class MenuJsonSchema(OpenAPISchema):
    """Matches MenuSchema but without the inherited AgentMessage properties."""

    title = fields.Str(
        required=False, metadata={"description": "Menu title", "example": "My Menu"}
    )
    description = fields.Str(
        required=False,
        metadata={
            "description": "Introductory text for the menu",
            "example": "User preferences for window settings",
        },
    )
    errormsg = fields.Str(
        required=False,
        metadata={
            "description": "Optional error message to display in menu header",
            "example": "Error: item not present",
        },
    )
    options = fields.List(
        fields.Nested(MenuOptionSchema),
        required=True,
        metadata={"description": "List of menu options"},
    )


class SendMenuSchema(OpenAPISchema):
    """Request schema for sending a menu to a connection."""

    menu = fields.Nested(
        MenuJsonSchema(),
        required=True,
        metadata={"description": "Menu to send to connection"},
    )


class MenuConnIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking connection id."""

    conn_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )


class ActionMenuFetchResultSchema(OpenAPISchema):
    """Result schema for action-menu fetch."""

    result = fields.Nested(MenuSchema, metadata={"description": "Action menu"})


@docs(tags=["action-menu"], summary="Close the active menu associated with a connection")
@match_info_schema(MenuConnIdMatchInfoSchema())
@response_schema(ActionMenuModulesResultSchema(), 200, description="")
@tenant_authentication
async def actionmenu_close(request: web.BaseRequest):
    """Request handler for closing the menu associated with a connection.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    connection_id = request.match_info["conn_id"]

    menu = await retrieve_connection_menu(connection_id, context)
    if not menu:
        raise web.HTTPNotFound(
            reason=f"No {MENU_RECORD_TYPE} record found for connection {connection_id}"
        )

    try:
        await save_connection_menu(None, connection_id, context)
    except StorageError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({})


@docs(tags=["action-menu"], summary="Fetch the active menu")
@match_info_schema(MenuConnIdMatchInfoSchema())
@response_schema(ActionMenuFetchResultSchema(), 200, description="")
@tenant_authentication
async def actionmenu_fetch(request: web.BaseRequest):
    """Request handler for fetching the previously-received menu for a connection.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    connection_id = request.match_info["conn_id"]

    menu = await retrieve_connection_menu(connection_id, context)
    result = {"result": menu.serialize() if menu else None}
    return web.json_response(result)


@docs(tags=["action-menu"], summary="Perform an action associated with the active menu")
@match_info_schema(MenuConnIdMatchInfoSchema())
@request_schema(PerformRequestSchema())
@response_schema(ActionMenuModulesResultSchema(), 200, description="")
@tenant_authentication
async def actionmenu_perform(request: web.BaseRequest):
    """Request handler for performing a menu action.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    connection_id = request.match_info["conn_id"]
    outbound_handler = request["outbound_message_router"]
    params = await request.json()

    try:
        async with context.profile.session() as session:
            connection = await ConnRecord.retrieve_by_id(session, connection_id)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    if connection.is_ready:
        msg = Perform(name=params["name"], params=params.get("params"))
        await outbound_handler(msg, connection_id=connection_id)
        return web.json_response({})

    raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")


@docs(tags=["action-menu"], summary="Request the active menu")
@match_info_schema(MenuConnIdMatchInfoSchema())
@response_schema(ActionMenuModulesResultSchema(), 200, description="")
@tenant_authentication
async def actionmenu_request(request: web.BaseRequest):
    """Request handler for requesting a menu from the connection target.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    connection_id = request.match_info["conn_id"]
    outbound_handler = request["outbound_message_router"]

    try:
        async with context.profile.session() as session:
            connection = await ConnRecord.retrieve_by_id(session, connection_id)
    except StorageNotFoundError as err:
        LOGGER.debug("Connection not found for action menu request: %s", connection_id)
        raise web.HTTPNotFound(reason=err.roll_up) from err

    if connection.is_ready:
        msg = MenuRequest()
        await outbound_handler(msg, connection_id=connection_id)
        return web.json_response({})

    raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")


@docs(tags=["action-menu"], summary="Send an action menu to a connection")
@match_info_schema(MenuConnIdMatchInfoSchema())
@request_schema(SendMenuSchema())
@response_schema(ActionMenuModulesResultSchema(), 200, description="")
@tenant_authentication
async def actionmenu_send(request: web.BaseRequest):
    """Request handler for requesting a menu from the connection target.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    connection_id = request.match_info["conn_id"]
    outbound_handler = request["outbound_message_router"]
    menu_json = await request.json()
    LOGGER.debug("Received send-menu request: %s %s", connection_id, menu_json)
    try:
        msg = Menu.deserialize(menu_json["menu"])
    except BaseModelError as err:
        LOGGER.exception("Exception deserializing menu: %s", err.roll_up)
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    try:
        async with context.profile.session() as session:
            connection = await ConnRecord.retrieve_by_id(session, connection_id)
    except StorageNotFoundError as err:
        LOGGER.debug(
            "Connection not found for action menu send request: %s", connection_id
        )
        raise web.HTTPNotFound(reason=err.roll_up) from err

    if connection.is_ready:
        await outbound_handler(msg, connection_id=connection_id)
        return web.json_response({})

    raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.post("/action-menu/{conn_id}/close", actionmenu_close),
            web.post("/action-menu/{conn_id}/fetch", actionmenu_fetch),
            web.post("/action-menu/{conn_id}/perform", actionmenu_perform),
            web.post("/action-menu/{conn_id}/request", actionmenu_request),
            web.post("/action-menu/{conn_id}/send-menu", actionmenu_send),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {"name": "action-menu", "description": "Menu interaction over connection"}
    )



"""Connection handling admin routes."""

import json

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import fields, validate

from ....admin.decorators.auth import tenant_authentication
from ....admin.request_context import AdminRequestContext
from ....connections.models.conn_record import ConnRecord, ConnRecordSchema
from ....messaging.models.base import BaseModelError
from ....messaging.models.openapi import OpenAPISchema
from ....messaging.valid import (
    ENDPOINT_EXAMPLE,
    ENDPOINT_VALIDATE,
    GENERIC_DID_EXAMPLE,
    GENERIC_DID_VALIDATE,
    UUID4_EXAMPLE,
    UUID4_VALIDATE,
)
from ....storage.error import StorageError, StorageNotFoundError
from ....wallet.base import BaseWallet
from ....wallet.error import WalletError
from .manager import DIDXManager, DIDXManagerError
from .message_types import DIDEX_1_0, DIDEX_1_1, SPEC_URI
from .messages.request import DIDXRequest, DIDXRequestSchema


class DIDXAcceptInvitationQueryStringSchema(OpenAPISchema):
    """Parameters and validators for accept invitation request query string."""

    my_endpoint = fields.Str(
        required=False,
        validate=ENDPOINT_VALIDATE,
        metadata={"description": "My URL endpoint", "example": ENDPOINT_EXAMPLE},
    )
    my_label = fields.Str(
        required=False,
        metadata={"description": "Label for connection request", "example": "Broker"},
    )
    use_did = fields.Str(
        required=False,
        metadata={
            "description": "The DID to use to for this connection",
            "example": "did:example:1234",
        },
    )
    use_did_method = fields.Str(
        required=False,
        validate=validate.OneOf(DIDXManager.SUPPORTED_USE_DID_METHODS),
        metadata={
            "description": "The DID method to use to generate a DID for this connection",
            "example": "did:peer:4",
        },
    )


class DIDXCreateRequestImplicitQueryStringSchema(OpenAPISchema):
    """Parameters and validators for create-request-implicit request query string."""

    their_public_did = fields.Str(
        required=True,
        allow_none=False,
        validate=GENERIC_DID_VALIDATE,
        metadata={
            "description": "Qualified public DID to which to request connection",
            "example": GENERIC_DID_EXAMPLE,
        },
    )
    alias = fields.Str(
        required=False,
        metadata={"description": "Alias for connection", "example": "Barry"},
    )
    auto_accept = fields.Boolean(
        required=False,
        metadata={"description": "Auto-accept connection (defaults to configuration)"},
    )
    my_endpoint = fields.Str(
        required=False,
        validate=ENDPOINT_VALIDATE,
        metadata={"description": "My URL endpoint", "example": ENDPOINT_EXAMPLE},
    )
    my_label = fields.Str(
        required=False,
        metadata={"description": "Label for connection request", "example": "Broker"},
    )
    mediation_id = fields.Str(
        required=False,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Identifier for active mediation record to be used",
            "example": UUID4_EXAMPLE,
        },
    )
    use_public_did = fields.Boolean(
        required=False, metadata={"description": "Use public DID for this connection"}
    )
    use_did = fields.Str(
        required=False,
        metadata={
            "description": "The DID to use to for this connection",
            "example": "did:example:1234",
        },
    )
    use_did_method = fields.Str(
        required=False,
        validate=validate.OneOf(DIDXManager.SUPPORTED_USE_DID_METHODS),
        metadata={
            "description": "The DID method to use to generate a DID for this connection",
            "example": "did:peer:4",
        },
    )
    goal_code = fields.Str(
        required=False,
        metadata={
            "description": (
                "A self-attested code the receiver may want to display to the user or"
                " use in automatically deciding what to do with the out-of-band message"
            ),
            "example": "issue-vc",
        },
    )
    goal = fields.Str(
        required=False,
        metadata={
            "description": (
                "A self-attested string that the receiver may want to display to the"
                " user about the context-specific goal of the out-of-band message"
            ),
            "example": "To issue a Faber College Graduate credential",
        },
    )
    protocol = fields.Str(
        required=False,
        validate=validate.OneOf([DIDEX_1_0, DIDEX_1_1]),
        metadata={
            "description": "Which DID Exchange Protocol version to use",
            "example": "didexchange/1.0",
        },
    )


class DIDXReceiveRequestImplicitQueryStringSchema(OpenAPISchema):
    """Parameters and validators for receive-request-implicit request query string."""

    alias = fields.Str(
        required=False,
        metadata={"description": "Alias for connection", "example": "Barry"},
    )
    my_endpoint = fields.Str(
        required=False,
        validate=ENDPOINT_VALIDATE,
        metadata={"description": "My URL endpoint", "example": ENDPOINT_EXAMPLE},
    )
    auto_accept = fields.Boolean(
        required=False,
        metadata={"description": "Auto-accept connection (defaults to configuration)"},
    )
    mediation_id = fields.Str(
        required=False,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Identifier for active mediation record to be used",
            "example": UUID4_EXAMPLE,
        },
    )


class DIDXAcceptRequestQueryStringSchema(OpenAPISchema):
    """Parameters and validators for accept-request request query string."""

    my_endpoint = fields.Str(
        required=False,
        validate=ENDPOINT_VALIDATE,
        metadata={"description": "My URL endpoint", "example": ENDPOINT_EXAMPLE},
    )
    mediation_id = fields.Str(
        required=False,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Identifier for active mediation record to be used",
            "example": UUID4_EXAMPLE,
        },
    )
    use_public_did = fields.Boolean(
        required=False, metadata={"description": "Use public DID for this connection"}
    )


class DIDXConnIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking connection id."""

    conn_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )


class DIDXConnIdRefIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking connection and ref ids."""

    conn_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )

    ref_id = fields.Str(
        required=True,
        metadata={
            "description": "Inbound connection identifier",
            "example": UUID4_EXAMPLE,
        },
    )


class DIDXRejectRequestSchema(OpenAPISchema):
    """Parameters and validators for reject-request request string."""

    reason = fields.Str(
        metadata={
            "description": "Reason for rejecting the DID Exchange",
            "example": "Request rejected",
        },
        required=False,
    )


@docs(
    tags=["did-exchange"],
    summary="Accept a stored connection invitation",
)
@match_info_schema(DIDXConnIdMatchInfoSchema())
@querystring_schema(DIDXAcceptInvitationQueryStringSchema())
@response_schema(ConnRecordSchema(), 200, description="")
@tenant_authentication
async def didx_accept_invitation(request: web.BaseRequest):
    """Request handler for accepting a stored connection invitation.

    Args:
        request: aiohttp request object

    Returns:
        The resulting connection record details

    """
    context: AdminRequestContext = request["context"]

    outbound_handler = request["outbound_message_router"]
    connection_id = request.match_info["conn_id"]
    my_label = request.query.get("my_label") or None
    my_endpoint = request.query.get("my_endpoint") or None
    mediation_id = request.query.get("mediation_id") or None
    use_did = request.query.get("use_did") or None
    use_did_method = request.query.get("use_did_method") or None

    if use_did and use_did_method:
        raise web.HTTPBadRequest(
            reason="use_did and use_did_method are mutually exclusive"
        )

    profile = context.profile
    didx_mgr = DIDXManager(profile)
    try:
        async with profile.session() as session:
            conn_rec = await ConnRecord.retrieve_by_id(session, connection_id)
            if use_did:
                wallet = session.inject(BaseWallet)
                did_info = await wallet.get_local_did(use_did)
                conn_rec.my_did = did_info.did
                await conn_rec.save(
                    session, reason="Set my_did from use_did on invite accept"
                )

        didx_request = await didx_mgr.create_request(
            conn_rec=conn_rec,
            my_label=my_label,
            my_endpoint=my_endpoint,
            mediation_id=mediation_id,
            use_did_method=use_did_method,
        )
        result = conn_rec.serialize()
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (StorageError, WalletError, DIDXManagerError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_handler(didx_request, connection_id=conn_rec.connection_id)

    return web.json_response(result)


@docs(
    tags=["did-exchange"],
    summary="Create and send a request against public DID's implicit invitation",
)
@querystring_schema(DIDXCreateRequestImplicitQueryStringSchema())
@response_schema(ConnRecordSchema(), 200, description="")
@tenant_authentication
async def didx_create_request_implicit(request: web.BaseRequest):
    """Request handler for creating and sending a request to an implicit invitation.

    Args:
        request: aiohttp request object

    Returns:
        The resulting connection record details

    """
    context: AdminRequestContext = request["context"]

    their_public_did = request.query["their_public_did"]
    my_label = request.query.get("my_label") or None
    my_endpoint = request.query.get("my_endpoint") or None
    mediation_id = request.query.get("mediation_id") or None
    alias = request.query.get("alias") or None
    use_public_did = json.loads(request.query.get("use_public_did", "null"))
    use_did = request.query.get("use_did") or None
    use_did_method = request.query.get("use_did_method") or None
    goal_code = request.query.get("goal_code") or None
    goal = request.query.get("goal") or None
    auto_accept = json.loads(request.query.get("auto_accept", "null"))
    protocol = request.query.get("protocol") or None

    profile = context.profile
    didx_mgr = DIDXManager(profile)
    try:
        didx_request = await didx_mgr.create_request_implicit(
            their_public_did=their_public_did,
            my_label=my_label,
            my_endpoint=my_endpoint,
            mediation_id=mediation_id,
            use_public_did=use_public_did,
            use_did=use_did,
            use_did_method=use_did_method,
            alias=alias,
            goal_code=goal_code,
            goal=goal,
            auto_accept=auto_accept,
            protocol=protocol,
        )
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (StorageError, WalletError, DIDXManagerError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(didx_request.serialize())


@docs(
    tags=["did-exchange"],
    summary="Receive request against public DID's implicit invitation",
    deprecated=True,
)
@querystring_schema(DIDXReceiveRequestImplicitQueryStringSchema())
@request_schema(DIDXRequestSchema())
@response_schema(ConnRecordSchema(), 200, description="")
@tenant_authentication
async def didx_receive_request_implicit(request: web.BaseRequest):
    """Request handler for receiving a request against public DID's implicit invitation.

    Args:
        request: aiohttp request object

    Returns:
        The resulting connection record details

    """
    context: AdminRequestContext = request["context"]

    body = await request.json()
    alias = request.query.get("alias")
    auto_accept = json.loads(request.query.get("auto_accept", "null"))

    profile = context.profile
    didx_mgr = DIDXManager(profile)
    try:
        didx_request = DIDXRequest.deserialize(body)
        conn_rec = await didx_mgr.receive_request(
            request=didx_request,
            recipient_did=didx_request._thread.pthid.split(":")[-1],
            alias=alias,
            auto_accept_implicit=auto_accept,
        )
        result = conn_rec.serialize()
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (StorageError, WalletError, DIDXManagerError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(result)


@docs(
    tags=["did-exchange"],
    summary="Accept a stored connection request",
)
@match_info_schema(DIDXConnIdMatchInfoSchema())
@querystring_schema(DIDXAcceptRequestQueryStringSchema())
@response_schema(ConnRecordSchema(), 200, description="")
@tenant_authentication
async def didx_accept_request(request: web.BaseRequest):
    """Request handler for accepting a stored connection request.

    Args:
        request: aiohttp request object

    Returns:
        The resulting connection record details

    """
    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]
    connection_id = request.match_info["conn_id"]
    my_endpoint = request.query.get("my_endpoint") or None
    mediation_id = request.query.get("mediation_id") or None
    use_public_did = json.loads(request.query.get("use_public_did", "null"))

    profile = context.profile
    didx_mgr = DIDXManager(profile)
    try:
        async with profile.session() as session:
            conn_rec = await ConnRecord.retrieve_by_id(session, connection_id)
        response = await didx_mgr.create_response(
            conn_rec=conn_rec,
            my_endpoint=my_endpoint,
            mediation_id=mediation_id,
            use_public_did=use_public_did,
        )
        result = conn_rec.serialize()
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (StorageError, WalletError, DIDXManagerError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_handler(response, connection_id=conn_rec.connection_id)
    return web.json_response(result)


@docs(
    tags=["did-exchange"],
    summary="Abandon or reject a DID Exchange",
)
@match_info_schema(DIDXConnIdMatchInfoSchema())
@request_schema(DIDXRejectRequestSchema())
@response_schema(ConnRecordSchema(), 200, description="")
@tenant_authentication
async def didx_reject(request: web.BaseRequest):
    """Abandon or reject a DID Exchange."""
    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    connection_id = request.match_info["conn_id"]
    body = await request.json()
    reason = body.get("reason")

    profile = context.profile
    didx_mgr = DIDXManager(profile)
    try:
        async with profile.session() as session:
            conn_rec = await ConnRecord.retrieve_by_id(session, connection_id)
        report = await didx_mgr.reject(conn_rec=conn_rec, reason=reason)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (StorageError, WalletError, DIDXManagerError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_handler(report, connection_id=conn_rec.connection_id)
    result = conn_rec.serialize()
    return web.json_response(result)


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.post(
                "/didexchange/{conn_id}/accept-invitation",
                didx_accept_invitation,
            ),
            web.post("/didexchange/create-request", didx_create_request_implicit),
            web.post("/didexchange/receive-request", didx_receive_request_implicit),
            web.post("/didexchange/{conn_id}/accept-request", didx_accept_request),
            web.post("/didexchange/{conn_id}/reject", didx_reject),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "did-exchange",
            "description": "Connection management via DID exchange",
            "externalDocs": {"description": "Specification", "url": SPEC_URI},
        }
    )



"""WEBVH DID Resolver.

Resolution is performed by the did_webvh library.
"""

from re import Pattern
from typing import Optional, Sequence, Text

from did_webvh.resolver import ResolutionResult, resolve_did

from ...config.injection_context import InjectionContext
from ...core.profile import Profile
from ...messaging.valid import DIDWebvh
from ..base import BaseDIDResolver, ResolverType


class WebvhDIDResolver(BaseDIDResolver):
    """WEBVH DID Resolver."""

    def __init__(self):
        """Initialize the WEBVH DID Resolver."""
        super().__init__(ResolverType.NATIVE)

    async def setup(self, context: InjectionContext):
        """Perform required setup for WEBVH DID resolution."""

    @property
    def supported_did_regex(self) -> Pattern:
        """Return supported DID regex of WEBVH DID Resolver."""
        return DIDWebvh.PATTERN

    async def _resolve(
        self, profile: Profile, did: str, service_accept: Optional[Sequence[Text]] = None
    ) -> dict:
        """Resolve DID using WEBVH."""
        response: ResolutionResult = await resolve_did(did)
        if response.resolution_metadata and response.resolution_metadata.get("error"):
            return response.resolution_metadata

        return response.document



"""Test mediate grant message handler."""

import pytest
import pytest_asyncio

from acapy_agent.core.profile import ProfileSession
from acapy_agent.tests import mock

from ......connections.models.conn_record import ConnRecord
from ......messaging.base_handler import HandlerException
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......multitenant.base import BaseMultitenantManager
from ......utils.testing import create_test_profile
from ...manager import MediationManager
from ...messages.mediate_grant import MediationGrant
from ...models.mediation_record import MediationRecord
from .. import mediation_grant_handler as test_module
from ..mediation_grant_handler import MediationGrantHandler

TEST_CONN_ID = "conn-id"
TEST_BASE58_VERKEY = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
TEST_VERKEY = "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
TEST_ENDPOINT = "https://example.com"


@pytest_asyncio.fixture
async def context():
    context = RequestContext.test_context(await create_test_profile())
    context.message = MediationGrant(endpoint=TEST_ENDPOINT, routing_keys=[TEST_VERKEY])
    context.connection_ready = True
    context.connection_record = ConnRecord(connection_id=TEST_CONN_ID)
    yield context


@pytest_asyncio.fixture
async def session(context: RequestContext):
    yield await context.session()


@pytest.mark.asyncio
class TestMediationGrantHandler:
    """Test mediate grant message handler."""

    async def test_handler_no_active_connection(self, context: RequestContext):
        handler, responder = MediationGrantHandler(), MockResponder()
        context.connection_ready = False
        with pytest.raises(HandlerException):
            await handler.handle(context, responder)

    async def test_handler_no_mediation_record(self, context: RequestContext):
        handler, responder = MediationGrantHandler(), MockResponder()
        with pytest.raises(HandlerException):
            await handler.handle(context, responder)

    @pytest.mark.parametrize(
        "grant",
        [
            MediationGrant(endpoint=TEST_ENDPOINT, routing_keys=[TEST_VERKEY]),
            MediationGrant(endpoint=TEST_ENDPOINT, routing_keys=[TEST_BASE58_VERKEY]),
        ],
    )
    async def test_handler(
        self, grant: MediationGrant, session: ProfileSession, context: RequestContext
    ):
        handler, responder = MediationGrantHandler(), MockResponder()
        await MediationRecord(connection_id=TEST_CONN_ID).save(session)
        context.message = grant
        await handler.handle(context, responder)
        record = await MediationRecord.retrieve_by_connection_id(session, TEST_CONN_ID)
        assert record
        assert record.state == MediationRecord.STATE_GRANTED
        assert record.endpoint == TEST_ENDPOINT
        assert record.routing_keys == [TEST_VERKEY]

    async def test_handler_connection_has_set_to_default_meta(
        self, session: ProfileSession, context: RequestContext
    ):
        handler, responder = MediationGrantHandler(), MockResponder()
        record = MediationRecord(connection_id=TEST_CONN_ID)
        await record.save(session)
        with (
            mock.patch.object(
                context.connection_record,
                "metadata_get",
                mock.CoroutineMock(return_value=True),
            ),
            mock.patch.object(
                test_module, "MediationManager", autospec=True
            ) as mock_mediation_manager,
        ):
            await handler.handle(context, responder)
            mock_mediation_manager.return_value.set_default_mediator.assert_called_once_with(
                record
            )

    async def test_handler_multitenant_base_mediation(
        self, session: ProfileSession, context: RequestContext
    ):
        handler, responder = MediationGrantHandler(), mock.CoroutineMock()
        responder.send = mock.CoroutineMock()
        profile = context.profile

        profile.context.update_settings(
            {"multitenant.enabled": True, "wallet.id": "test_wallet"}
        )

        multitenant_mgr = mock.MagicMock(BaseMultitenantManager, autospec=True)

        default_base_mediator = MediationRecord(routing_keys=["key1", "key2"])
        multitenant_mgr.get_default_mediator = mock.CoroutineMock()
        multitenant_mgr.get_default_mediator.return_value = default_base_mediator
        profile.context.injector.bind_instance(BaseMultitenantManager, multitenant_mgr)

        record = MediationRecord(connection_id=TEST_CONN_ID)
        await record.save(session)
        with mock.patch.object(MediationManager, "add_key") as add_key:
            keylist_updates = mock.MagicMock()
            add_key.return_value = keylist_updates

            await handler.handle(context, responder)

            add_key.assert_called_once_with("key2")
            responder.send.assert_called_once_with(
                keylist_updates, connection_id=TEST_CONN_ID
            )

    async def test_handler_connection_no_set_to_default(
        self, session: ProfileSession, context: RequestContext
    ):
        handler, responder = MediationGrantHandler(), MockResponder()
        record = MediationRecord(connection_id=TEST_CONN_ID)
        await record.save(session)
        with (
            mock.patch.object(
                context.connection_record,
                "metadata_get",
                mock.CoroutineMock(return_value=False),
            ),
            mock.patch.object(
                test_module, "MediationManager", autospec=True
            ) as mock_mediation_manager,
        ):
            await handler.handle(context, responder)
            mock_mediation_manager.return_value.set_default_mediator.assert_not_called()



from datetime import datetime
from unittest import IsolatedAsyncioTestCase, mock

from ....did.did_key import DIDKey
from ....utils.testing import create_test_profile
from ....wallet.base import BaseWallet
from ....wallet.key_type import ED25519
from ...ld_proofs import Ed25519Signature2020, WalletKeyPair
from ...ld_proofs.error import LinkedDataProofException
from ...tests.document_loader import custom_document_loader
from ...vc_ld import issue_vc as issue
from ...vc_ld import sign_presentation, verify_credential, verify_presentation
from .test_credential_v2 import (
    CREDENTIAL_V2_ISSUED,
    CREDENTIAL_V2_TEMPLATE,
    CREDENTIAL_V2_VERIFIED,
    PRESENTATION_V2_SIGNED,
    PRESENTATION_V2_UNSIGNED,
)


class TestLinkedDataVerifiableCredentialV2(IsolatedAsyncioTestCase):
    test_seed = "testseed000000000000000000000001"

    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            self.ed25519_key_info = await wallet.create_signing_key(
                key_type=ED25519, seed=self.test_seed
            )
            self.ed25519_verification_method = DIDKey.from_public_key_b58(
                self.ed25519_key_info.verkey, ED25519
            ).key_id

        self.presentation_challenge = "2b1bbff6-e608-4368-bf84-67471b27e41c"

    async def test_v2_issue_Ed25519Signature2020(self):
        suite = Ed25519Signature2020(
            verification_method=self.ed25519_verification_method,
            key_pair=WalletKeyPair(
                profile=self.profile,
                key_type=ED25519,
                public_key_base58=self.ed25519_key_info.verkey,
            ),
            date=datetime.strptime(
                "2025-01-27T21:39:18+00:00", "%Y-%m-%dT%H:%M:%S+00:00"
            ),
        )

        issued = await issue(
            credential=CREDENTIAL_V2_TEMPLATE,
            suite=suite,
            document_loader=custom_document_loader,
        )
        assert issued == CREDENTIAL_V2_ISSUED

    async def test_v2_verify_Ed25519Signature2020(self):
        # Verification requires lot less input parameters
        suite = Ed25519Signature2020(
            key_pair=WalletKeyPair(profile=self.profile, key_type=ED25519),
        )
        verified = await verify_credential(
            credential=CREDENTIAL_V2_ISSUED,
            suites=[suite],
            document_loader=custom_document_loader,
        )

        assert verified == CREDENTIAL_V2_VERIFIED

    async def test_v2_verify_presentation(self):
        suite = Ed25519Signature2020(
            key_pair=WalletKeyPair(profile=self.profile, key_type=ED25519),
        )
        verification_result = await verify_presentation(
            presentation=PRESENTATION_V2_SIGNED,
            challenge=self.presentation_challenge,
            suites=[suite],
            document_loader=custom_document_loader,
        )

        assert verification_result.verified

    async def test_verify_presentation_x_no_purpose_challenge(self):
        verification_result = await verify_presentation(
            presentation=PRESENTATION_V2_SIGNED,
            suites=[],
            document_loader=custom_document_loader,
        )

        assert not verification_result.verified
        assert 'A "challenge" param is required for AuthenticationProofPurpose' in str(
            verification_result.errors[0]
        )

    async def test_sign_presentation_x_no_purpose_challenge(self):
        with self.assertRaises(LinkedDataProofException) as context:
            await sign_presentation(
                presentation=PRESENTATION_V2_UNSIGNED,
                suite=mock.MagicMock(),
                document_loader=mock.MagicMock(),
            )
        assert 'A "challenge" param is required' in str(context.exception)



"""Routes for revocation notification."""

import logging
import re

from ....core.event_bus import Event, EventBus
from ....core.profile import Profile
from ....messaging.responder import BaseResponder
from ....revocation.util import (
    REVOCATION_CLEAR_PENDING_EVENT,
    REVOCATION_EVENT_PREFIX,
    REVOCATION_PUBLISHED_EVENT,
)
from ....storage.error import StorageError, StorageNotFoundError
from .models.rev_notification_record import RevNotificationRecord

LOGGER = logging.getLogger(__name__)


def register_events(event_bus: EventBus):
    """Register to handle events."""
    event_bus.subscribe(
        re.compile(f"^{REVOCATION_EVENT_PREFIX}{REVOCATION_PUBLISHED_EVENT}.*"),
        on_revocation_published,
    )
    event_bus.subscribe(
        re.compile(f"^{REVOCATION_EVENT_PREFIX}{REVOCATION_CLEAR_PENDING_EVENT}.*"),
        on_pending_cleared,
    )


async def on_revocation_published(profile: Profile, event: Event):
    """Handle issuer revoke event."""
    LOGGER.debug("Received notification of revocation publication: %s", event.payload)

    should_notify = profile.settings.get("revocation.notify", False)
    responder = profile.inject(BaseResponder)
    crids = event.payload.get("crids") or []
    # Allow for crids to be integers
    if crids and isinstance(crids[0], int):
        crids = [str(crid) for crid in crids]

    try:
        async with profile.session() as session:
            records = await RevNotificationRecord.query_by_rev_reg_id(
                session,
                rev_reg_id=event.payload["rev_reg_id"],
            )
            records = [record for record in records if record.cred_rev_id in crids]

            for record in records:
                await record.delete_record(session)
                if should_notify:
                    await responder.send(
                        record.to_message(), connection_id=record.connection_id
                    )
                    LOGGER.info(
                        "Sent revocation notification for credential to %s",
                        record.connection_id,
                    )

    except StorageNotFoundError:
        LOGGER.info(
            "No revocation notification record found for revoked credential; "
            "no notification will be sent"
        )
    except StorageError:
        LOGGER.exception("Failed to retrieve revocation notification record")


async def on_pending_cleared(profile: Profile, event: Event):
    """Handle pending cleared event."""

    # Query by rev reg ID
    async with profile.session() as session:
        notifications = await RevNotificationRecord.query_by_rev_reg_id(
            session, event.payload["rev_reg_id"]
        )

    # Delete
    async with profile.transaction() as txn:
        for notification in notifications:
            await notification.delete_record(txn)
        await txn.commit()



"""Package-wide code and data."""

TEST_DID = "LjgpST2rjsoxYegQDRm7EL"
SCHEMA_NAME = "bc-reg"
SCHEMA_TXN = 12
SCHEMA_ID = f"{TEST_DID}:2:{SCHEMA_NAME}:1.0"
SCHEMA = {
    "ver": "1.0",
    "id": SCHEMA_ID,
    "name": SCHEMA_NAME,
    "version": "1.0",
    "attrNames": ["legalName", "jurisdictionId", "incorporationDate"],
    "seqNo": SCHEMA_TXN,
}
CRED_DEF_ID = f"{TEST_DID}:3:CL:12:tag1"
CRED_DEF = {
    "ver": "1.0",
    "id": CRED_DEF_ID,
    "schemaId": SCHEMA_TXN,
    "type": "CL",
    "tag": "tag1",
    "value": {
        "primary": {
            "n": "...",
            "s": "...",
            "r": {
                "master_secret": "...",
                "legalName": "...",
                "busId": "...",
                "jurisdictionId": "...",
                "incorporationDate": "...",
                "pic": "...",
            },
            "rctxt": "...",
            "z": "...",
        },
        "revocation": {
            "g": "1 ...",
            "g_dash": "1 ...",
            "h": "1 ...",
            "h0": "1 ...",
            "h1": "1 ...",
            "h2": "1 ...",
            "htilde": "1 ...",
            "h_cap": "1 ...",
            "u": "1 ...",
            "pk": "1 ...",
            "y": "1 ...",
        },
    },
}
REV_REG_DEF_TYPE = "CL_ACCUM"
REV_REG_ID = f"{TEST_DID}:4:{CRED_DEF_ID}:{REV_REG_DEF_TYPE}:tag1"
TAILS_DIR = "/tmp/indy/revocation/tails_files"
TAILS_HASH = "8UW1Sz5cqoUnK9hqQk7nvtKK65t7Chu3ui866J23sFyJ"
TAILS_LOCAL = f"{TAILS_DIR}/{TAILS_HASH}"
REV_REG_DEF = {
    "ver": "1.0",
    "id": REV_REG_ID,
    "revocDefType": "CL_ACCUM",
    "tag": "tag1",
    "credDefId": CRED_DEF_ID,
    "value": {
        "issuanceType": "ISSUANCE_ON_DEMAND",
        "maxCredNum": 5,
        "publicKeys": {"accumKey": {"z": "1 ..."}},
        "tailsHash": TAILS_HASH,
        "tailsLocation": TAILS_LOCAL,
    },
}
INDY_OFFER = {
    "schema_id": SCHEMA_ID,
    "cred_def_id": CRED_DEF_ID,
    "key_correctness_proof": {
        "c": "123467890",
        "xz_cap": "12345678901234567890",
        "xr_cap": [
            [
                "remainder",
                "1234567890",
            ],
            [
                "number",
                "12345678901234",
            ],
            [
                "master_secret",
                "12345678901234",
            ],
        ],
    },
    "nonce": "1234567890",
}
INDY_CRED_REQ = {
    "prover_did": TEST_DID,
    "cred_def_id": CRED_DEF_ID,
    "blinded_ms": {
        "u": "12345",
        "ur": "1 123467890ABCDEF",
        "hidden_attributes": ["master_secret"],
        "committed_attributes": {},
    },
    "blinded_ms_correctness_proof": {
        "c": "77777",
        "v_dash_cap": "12345678901234567890",
        "m_caps": {"master_secret": "271283714"},
        "r_caps": {},
    },
    "nonce": "9876543210",
}
INDY_CRED = {
    "schema_id": SCHEMA_ID,
    "cred_def_id": CRED_DEF_ID,
    "rev_reg_id": REV_REG_ID,
    "values": {
        "legalName": {
            "raw": "The Original House of Pies",
            "encoded": "108156129846915621348916581250742315326283968964",
        },
        "busId": {"raw": "11155555", "encoded": "11155555"},
        "jurisdictionId": {"raw": "1", "encoded": "1"},
        "incorporationDate": {
            "raw": "2021-01-01",
            "encoded": "121381685682968329568231",
        },
        "pic": {"raw": "cG90YXRv", "encoded": "125362825623562385689562"},
    },
    "signature": {
        "p_credential": {
            "m_2": "13683295623862356",
            "a": "1925723185621385238953",
            "e": "253516862326",
            "v": "26890295622385628356813632",
        },
        "r_credential": {
            "sigma": "1 00F81D",
            "c": "158698926BD09866E",
            "vr_prime_prime": "105682396DDF1A",
            "witness_signature": {"sigma_i": "1 ...", "u_i": "1 ...", "g_i": "1 ..."},
            "g_i": "1 ...",
            "i": 1,
            "m2": "862186285926592362384FA97FF3A4AB",
        },
    },
    "signature_correctness_proof": {
        "se": "10582965928638296868123",
        "c": "2816389562839651",
    },
    "rev_reg": {"accum": "21 ..."},
    "witness": {"omega": "21 ..."},
}

LD_PROOF_VC_DETAIL = {
    "credential": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        "type": ["VerifiableCredential", "UniversityDegreeCredential"],
        "credentialSubject": {"test": "key"},
        "issuanceDate": "2021-04-12",
        "issuer": f"did:sov:{TEST_DID}",
    },
    "options": {"proofType": "Ed25519Signature2018"},
}
LD_PROOF_VC = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
    ],
    "type": ["VerifiableCredential", "UniversityDegreeCredential"],
    "credentialSubject": {"test": "key"},
    "issuanceDate": "2021-04-12",
    "issuer": f"did:sov:{TEST_DID}",
    "proof": {
        "proofPurpose": "assertionMethod",
        "created": "2019-12-11T03:50:55",
        "type": "Ed25519Signature2018",
        "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
        "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..Q6amIrxGiSbM7Ce6DxlfwLCjVcYyclas8fMxaecspXFUcFW9DAAxKzgHx93FWktnlZjM_biitkMgZdStgvivAQ",
    },
}



"""Test did:web Resolver."""

import pytest

from ..web import WebDIDResolver


@pytest.fixture
def resolver():
    yield WebDIDResolver()


def test_transformation_domain_only(resolver):
    did = "did:web:example.com"
    url = resolver._WebDIDResolver__transform_to_url(did)
    assert url == "https://example.com/.well-known/did.json"


def test_transformation_domain_with_path(resolver):
    did = "did:web:example.com:department:example"
    url = resolver._WebDIDResolver__transform_to_url(did)
    assert url == "https://example.com/department/example/did.json"


def test_transformation_domain_with_port(resolver):
    did = "did:web:localhost%3A443"
    url = resolver._WebDIDResolver__transform_to_url(did)
    assert url == "https://localhost:443/.well-known/did.json"
